"""
検証関数の定義
"""

import sys
from pathlib import Path

# プロジェクトのルートディレクトリをPythonパスに追加
project_root = Path(__file__).resolve().parent.parent
if str(project_root) not in sys.path:
    sys.path.insert(0, str(project_root))

import numpy as np
import networkx as nx
from typing import Dict, List
from genice3.genice import GenIce3
from genice3.molecule import Molecule


def validate_basic_execution(genice: GenIce3, exporter, output: str) -> bool:
    """
    レベル1: 基本的な動作確認

    Args:
        genice: GenIce3インスタンス
        exporter: Exporterインスタンス
        output: 出力文字列

    Returns:
        bool: 検証が成功した場合True
    """
    # 出力が生成される
    assert output is not None, "Output should not be None"
    assert len(output) > 0, "Output should not be empty"

    return True


def validate_output_format(exporter_name: str, output: str) -> bool:
    """
    レベル2: 出力形式の検証

    Args:
        exporter_name: exporter名
        output: 出力文字列

    Returns:
        bool: 検証が成功した場合True
    """
    if exporter_name == "gromacs":
        lines = output.split("\n")
        # ヘッダー行が存在
        assert (
            "Generated by GenIce" in output
        ), "Gromacs header should contain 'Generated by GenIce'"
        # 最低限の行数がある
        assert (
            len(lines) >= 3
        ), f"Gromacs output should have at least 3 lines, got {len(lines)}"
        # 原子数行が存在し、整数としてパースできる
        try:
            atom_count = int(lines[1].strip())
            assert atom_count > 0, f"Atom count should be positive, got {atom_count}"
        except (ValueError, IndexError) as e:
            raise AssertionError(f"Failed to parse atom count from line 2: {e}")

    elif exporter_name == "cif":
        # CIF形式の必須キーワード
        assert "data_genice" in output, "CIF output should contain 'data_genice'"
        assert "_cell_length_a" in output, "CIF output should contain '_cell_length_a'"
        assert "loop_" in output, "CIF output should contain 'loop_'"
        assert (
            "_atom_site_label" in output
        ), "CIF output should contain '_atom_site_label'"

    return True


def validate_ice_rules(genice: GenIce3) -> bool:
    """
    レベル3: アイスルールの検証

    各水分子から2本の水素結合が出て、2本が入ることを確認

    Args:
        genice: GenIce3インスタンス

    Returns:
        bool: 検証が成功した場合True
    """
    digraph = genice.digraph

    # 各ノード（水分子）から2本の水素結合が出る
    for node in digraph.nodes():
        out_degree = digraph.out_degree(node)
        in_degree = digraph.in_degree(node)

        assert (
            out_degree == 2
        ), f"Node {node} has {out_degree} outgoing edges, expected 2"
        assert in_degree == 2, f"Node {node} has {in_degree} incoming edges, expected 2"

    return True


def validate_atom_count(genice: GenIce3, water_model: Molecule) -> bool:
    """
    レベル3: 原子数の検証

    Args:
        genice: GenIce3インスタンス
        water_model: 水分子モデル

    Returns:
        bool: 検証が成功した場合True
    """
    waters = genice.water_molecules(water_model=water_model)

    # 水分子数 × 分子あたりの原子数
    expected_atoms = len(waters) * len(water_model.sites)

    # 実際の原子数をカウント
    total_atoms = sum(len(mol.sites) for mol in waters.values())

    assert (
        total_atoms == expected_atoms
    ), f"Expected {expected_atoms} atoms, got {total_atoms}"

    return True


def validate_coordinates_in_cell(genice: GenIce3, water_model: Molecule) -> bool:
    """
    レベル3: 座標がセル内に収まっていることの検証

    Args:
        genice: GenIce3インスタンス
        water_model: 水分子モデル

    Returns:
        bool: 検証が成功した場合True
    """
    cell = genice.cell
    cell_inv = np.linalg.inv(cell)

    waters = genice.water_molecules(water_model=water_model)

    for mol in waters.values():
        for site in mol.sites:
            # 絶対座標を相対座標に変換
            rel_pos = site @ cell_inv
            # 0-1の範囲に正規化
            rel_pos = rel_pos - np.floor(rel_pos)

            # すべての成分が0-1の範囲内（少し余裕を持たせる）
            assert np.all(rel_pos >= -1e-10) and np.all(
                rel_pos <= 1 + 1e-10
            ), f"Coordinate {site} is outside cell (relative: {rel_pos})"

    return True


def validate_graph_structure(genice: GenIce3) -> bool:
    """
    レベル5: グラフ構造の検証

    Args:
        genice: GenIce3インスタンス

    Returns:
        bool: 検証が成功した場合True
    """
    graph = genice.graph
    digraph = genice.digraph

    # 無向グラフと有向グラフのノード数が一致
    assert (
        graph.number_of_nodes() == digraph.number_of_nodes()
    ), f"Graph node count mismatch: {graph.number_of_nodes()} vs {digraph.number_of_nodes()}"

    # 有向グラフが生成されている
    assert digraph.number_of_edges() > 0, "Digraph should have edges"

    # グラフが連結している（無向グラフとして）
    if isinstance(graph, nx.Graph):
        assert nx.is_connected(graph), "Graph should be connected"
    elif isinstance(graph, nx.DiGraph):
        # 有向グラフの場合は無向グラフに変換して確認
        undirected = graph.to_undirected()
        assert nx.is_connected(undirected), "Graph should be connected"

    return True


def validate_reproducibility(unitcell_name: str, seed: int = 1) -> bool:
    """
    レベル4: 再現性の検証

    同じseedで同じ結果が得られることを確認

    Args:
        unitcell_name: 単位胞名
        seed: 乱数シード

    Returns:
        bool: 検証が成功した場合True
    """
    from genice3.plugin import UnitCell, Exporter
    from io import StringIO

    # 1回目の実行
    genice1 = GenIce3(seed=seed)
    genice1.unitcell = UnitCell(unitcell_name)
    output_file1 = StringIO()
    Exporter("gromacs").dump(genice1, output_file1, water_model="4site")
    output1 = output_file1.getvalue()

    # 2回目の実行
    run2 = GenIce3(seed=seed)
    run2.unitcell = UnitCell(unitcell_name)
    output_file2 = StringIO()
    Exporter("gromacs").dump(run2, output_file2, water_model="4site")
    output2 = output_file2.getvalue()

    # 出力が完全に一致
    assert output1 == output2, "Results should be identical with same seed"

    return True


def validate_comprehensive(
    genice: GenIce3, exporter_name: str, output: str, water_model: Molecule
) -> bool:
    """
    総合的な検証

    Args:
        genice: GenIce3インスタンス
        exporter_name: exporter名
        output: 出力文字列
        water_model: 水分子モデル

    Returns:
        bool: すべての検証が成功した場合True
    """
    # レベル1: 基本動作
    validate_basic_execution(genice, None, output)

    # レベル2: 出力形式
    validate_output_format(exporter_name, output)

    # レベル3: 物理的整合性
    validate_ice_rules(genice)
    validate_atom_count(genice, water_model)
    validate_coordinates_in_cell(genice, water_model)

    # レベル5: 構造的整合性
    validate_graph_structure(genice)

    return True



"""
Data sources

"""

desc = {
    "ref": {"Ic": "Vos 1993"},
    "brief": "Cubic type of ice I.",
}

import genice3.unitcell
import numpy as np
import networkx as nx
from cif2ice import cellvectors


class UnitCell(genice3.unitcell.UnitCell):
    """
    ice1c単位胞を定義するクラス。
    """

    def __init__(self, **kwargs):
        pairs_str = """
        0 4
        0 5
        0 6
        0 7
        1 4
        1 5
        1 6
        1 7
        2 4
        2 5
        2 6
        2 7
        3 4
        3 5
        3 6
        3 7
        """.split(
            "\n"
        )
        pairs = []
        for line in pairs_str:
            cols = line.split()
            if len(cols) == 2:
                pairs.append((int(cols[0]), int(cols[1])))

        graph = nx.Graph(pairs)

        waters = np.fromstring(
            """
        0 0 0
        0.5 0.5 0
        0.5 0 0.5
        0 0.5 0.5
        0.25 0.25 0.25
        0.75 0.75 0.25
        0.75 0.25 0.75
        0.25 0.75 0.75
        """,
            sep=" ",
        ).reshape(-1, 3)

        coord = "relative"

        density = 0.92

        cell = cellvectors(a=4.0, b=4.0, c=4.0)

        super().__init__(
            cell=cell,
            lattice_sites=waters,
            graph=graph,
            coord=coord,
            density=density,
            **kwargs,
        )



desc = {
    "ref": {},
    "brief": "Most popular Ice I (hexagonal). "
    + "NOTE: Due to a historical reason, the crystal axes of hexagonal ice are exchanged. "
    + "If you want the basal plane to be Z axis, please use ice1h_unit instead.",
}

import genice3.unitcell
import numpy as np
from cif2ice import cellvectors


class UnitCell(genice3.unitcell.UnitCell):
    """
    ice1h単位胞を定義するクラス。
    """

    def __init__(self, **kwargs):
        graph = None  # pairsがない場合は自動生成

        waters = np.fromstring(
            """
        1.328 1.802 3.38
        5.267 4.524 1.109
        6.58 5.442 3.365
        5.267 4.542 5.629
        2.623 0.877 5.644
        2.667 5.488 5.625
        5.241 1.756 1.12
        5.241 1.774 5.64
        1.354 4.588 7.888
        1.354 4.57 3.369
        2.667 5.47 1.105
        2.623 0.858 1.124
        6.537 0.831 3.384
        6.537 0.849 7.903
        6.581 5.461 7.884
        1.328 1.82 7.899
        """,
            sep=" ",
        ).reshape(-1, 3)

        coord = "absolute"

        bondlen = 3.0

        density = 0.92

        cell = cellvectors(a=7.84813412606925, b=7.37735062301457, c=9.06573834219084)

        super().__init__(
            cell=cell,
            lattice_sites=waters,
            coord=coord,
            bondlen=bondlen,
            density=density,
            **kwargs,
        )



# Another plan of reactive GenIce3 (リアクティブプロパティによる実装).

from genice3 import ConfigurationError
from genice3.group import Group
from genice3.molecule import Molecule
from genice3.plugin import UnitCell as UnitCellPlugin, safe_import
from genice3.util import (
    replicate_positions,
    grandcell_wrap,
    assume_tetrahedral_vectors,
)
from genice3.cage import CageSpecs, CageSpec
from cif2ice import cellshape
import genice_core
import networkx as nx
import numpy as np
from dataclasses import dataclass
from logging import getLogger
from typing import Any, Dict, Generator, List, Tuple
from enum import Enum
import inspect

from dependency_engine import DependencyEngine, get_reactive_tasks, reactive
from genice3.unitcell import UnitCell


def _graph_degree_stats(g: nx.Graph) -> str:
    """グラフの次数分布（0,1,2,3,4,>4 のノード数と割合）を文字列で返す。"""
    n = g.number_of_nodes()
    if n == 0:
        return "nodes=0"
    deg = [g.degree(v) for v in g]
    buckets = {0: 0, 1: 0, 2: 0, 3: 0, 4: 0, ">4": 0}
    for d in deg:
        if d <= 4:
            buckets[d] += 1
        else:
            buckets[">4"] += 1
    parts = [f"deg{k}:{c}({100*c/n:.1f}%)" for k, c in buckets.items() if c]
    return f"nodes={n} " + " ".join(parts)


class ShowUsageError(Exception):
    """Usage表示を要求する例外

    Args:
        flag_name: フラグ名（例: "?", "help?", "cage?"）
    """

    def __init__(self, flag_name: str, message: str = ""):
        self.flag_name = flag_name
        super().__init__(message or f"Show usage for flag: {flag_name}")


# enumのようなものはどう定義する?
class MoleculeType(Enum):
    WATER = "water"
    GUEST = "guest"
    DOPANT = "dopant"
    GROUP = "group"


@dataclass
class GuestSpec:
    """
    ゲストの情報を表すデータクラス。
    """

    molecule: Molecule
    occupancy: float

    def __repr__(self) -> str:
        return (
            f"GuestSpec(molecule={self.molecule.name!r}, "
            f"occupancy={self.occupancy:.3f})"
        )


@dataclass
class AtomicStructure:
    """
    原子構造データを統合的に保持するデータクラス。
    exporterプラグインが使用するための統一インターフェース。
    """

    waters: Dict[int, Molecule]
    guests: List[Molecule]
    ions: Dict[int, Molecule]
    cell: np.ndarray

    def __repr__(self) -> str:
        return (
            f"AtomicStructure(n_waters={len(self.waters)}, "
            f"n_guests={len(self.guests)}, "
            f"n_ions={len(self.ions)}, "
            f"cell_shape={self.cell.shape})"
        )


def _assume_water_orientations(
    coord: np.ndarray, digraph: nx.DiGraph, cellmat: np.ndarray, dopants: Dict[int, str]
) -> np.ndarray:
    """有向グラフと座標から各水分子の配向行列（Nx3x3）を計算する。

    2本のOHが一直線上にある場合は正しく動作しない。

    Args:
        coord: 各ノードの分数座標（Nx3）。
        digraph: 水素結合の向きが決まった有向グラフ。
        cellmat: セル行列（分数→直交変換に使用）。
        dopants: ドーパントが占めるサイト（ノード番号→イオン名）。

    Returns:
        各水分子の配向行列（Nx3x3）。直交座標系での回転行列。
    """

    logger = getLogger()
    # just for a test of pure water
    if len(coord) != digraph.number_of_nodes():
        raise ValueError(
            f"coord length ({len(coord)}) must match digraph node count "
            f"({digraph.number_of_nodes()})"
        )
    if len(dopants):
        logger.info(f"  {dopants} dopants")
    # 通常の氷であればアルゴリズムを高速化できる。

    nnode = len(list(digraph))
    neis = np.zeros([nnode, 2], dtype=int)

    # 仮想ノード用の配列。第0要素は実際には第nnode要素を表す。
    extended_coord = []

    celli = np.linalg.inv(cellmat)
    # v0 = np.zeros([nnode, 3])
    # v1 = np.zeros([nnode, 3])
    for node in digraph:
        if node in dopants:
            h1 = np.array([0.0, 1, 1]) / (2**0.5)
            h2 = np.array([0.0, -1, 1]) / (2**0.5)
            r1 = h1 @ celli
            r2 = h2 @ celli
            # 仮想ノードにさしかえる
            neis[node] = [nnode + len(extended_coord), nnode + len(extended_coord) + 1]
            extended_coord += [coord[node] + r1, coord[node] + r2]
            continue
        succ = list(digraph.successors(node))
        if len(succ) < 2:
            vsucc = (coord[succ] - coord[node]) @ cellmat
            pred = list(digraph.predecessors(node))
            vpred = (coord[pred] - coord[node]) @ cellmat
            vsucc /= np.linalg.norm(vsucc, axis=1)[:, np.newaxis]
            vpred /= np.linalg.norm(vpred, axis=1)[:, np.newaxis]
            if len(vpred) > 2:
                # number of incoming bonds should be <= 2
                vpred = vpred[:2]
            vcomp = assume_tetrahedral_vectors(np.vstack([vpred, vsucc]))
            logger.debug(f"Node {node} vcomp {vcomp} vsucc {vsucc} vpred {vpred}")
            vsucc = np.vstack([vsucc, vcomp])[:2]
            rsucc = vsucc @ celli
            # 仮想ノードにさしかえる
            neis[node] = [nnode + len(extended_coord), nnode + len(extended_coord) + 1]
            extended_coord += [coord[node] + rsucc[0], coord[node] + rsucc[1]]
        else:
            neis[node] = succ

    if len(extended_coord) == 0:
        extended_coord = coord
    else:
        extended_coord = np.vstack([coord, extended_coord])

    # array of donating vectors
    v0 = extended_coord[neis[:, 0]] - coord[:]
    v0 -= np.floor(v0 + 0.5)
    v0 = v0 @ cellmat
    v0 /= np.linalg.norm(v0, axis=1)[:, np.newaxis]
    v1 = extended_coord[neis[:, 1]] - coord[:]
    v1 -= np.floor(v1 + 0.5)
    v1 = v1 @ cellmat
    v1 /= np.linalg.norm(v1, axis=1)[:, np.newaxis]
    # intramolecular axes
    y = v1 - v0
    y /= np.linalg.norm(y, axis=1)[:, np.newaxis]
    z = v0 + v1
    z /= np.linalg.norm(z, axis=1)[:, np.newaxis]
    x = np.cross(y, z, axisa=1, axisb=1)

    rotmatrices = np.zeros([nnode, 3, 3])

    rotmatrices[:, 0, :] = x
    rotmatrices[:, 1, :] = y
    rotmatrices[:, 2, :] = z
    return rotmatrices


def _replicate_lattice_node(
    lattice_site_node: int, nmol: int, replication_matrix: np.ndarray
) -> Generator[int, None, None]:
    """1つの格子サイトを拡大単位胞内で複製したときのノード番号を列挙する。

    Args:
        lattice_site_node: 複製前の単位胞内での格子サイト（ノード）番号。
        nmol: 単位胞内の水分子数（ノード数）。
        replication_matrix: 拡大倍率（行列式）の計算に用いる3x3行列。通常は replication_matrix を渡す。

    Yields:
        拡大単位胞内でのノード番号（lattice_site_node + nmol * i for i in 0..倍率-1）。
    """
    multiple = int(np.floor(np.linalg.det(replication_matrix) + 0.5))
    for i in range(multiple):
        yield lattice_site_node + nmol * i


def _replicate_graph(
    graph1: nx.Graph,
    cell1frac_coords: np.ndarray,
    replica_vectors: np.ndarray,
    replica_vector_index: Dict[Tuple[int, ...], int],
    reshape: np.ndarray,
) -> nx.Graph:
    """
    指定されたレプリカベクトルと形状に基づいてグラフを複製する。

    2つの座標系がいりみだれているので注意。
    cell1frac: 複製前の単位胞における小数座標
    grandfrac: 複製後の大きな単位胞における小数座標

    Args:
      graph1: 元のグラフ。
      cell1frac_coords: 元のグラフ内の節点の分数座標 (Nx3)。
      replica_vectors: レプリカベクトルの配列。
      replica_vector_index: レプリカベクトル座標タプル → 一意のインデックス。
      reshape: 拡大単位胞の積み重ね方を表す行列。

    Returns:
      複製された無向グラフ。fixed の複製は行わない。
    """
    # repgraph = dg.IceGraph()
    repgraph = nx.Graph()
    nmol = cell1frac_coords.shape[0]

    # 正の行列式の値(倍率)。整数。
    det = np.linalg.det(reshape)
    if det < 0:
        det = -det
    det = np.floor(det + 0.5).astype(int)
    # 逆行列に行列式をかけたもの。整数行列。
    invdet = np.floor(np.linalg.inv(reshape) * det + 0.5).astype(int)

    for i, j in graph1.edges(data=False):
        # positions in the original small cell
        cell1_delta = cell1frac_coords[j] - cell1frac_coords[i]
        cell1_delta = np.floor(cell1_delta + 0.5).astype(int)

        for a, cell1frac_a in enumerate(replica_vectors):
            cell1frac_b = cell1frac_a + cell1_delta
            cell1frac_b = np.floor(
                grandcell_wrap(cell1frac_b, reshape, invdet, det)
            ).astype(int)
            b = replica_vector_index[tuple(cell1frac_b)]
            newi = nmol * b + i
            newj = nmol * a + j

            repgraph.add_edge(newi, newj)

    return repgraph


def _replicate_fixed_edges(
    repgraph: nx.Graph, fixed: nx.DiGraph, nmol: int
) -> nx.DiGraph:
    """単位胞の固定エッジを拡大単位胞のグラフ上に複製した有向グラフを返す。

    Args:
        repgraph: 拡大単位胞全体の無向グラフ。
        fixed: 単位胞内の固定エッジ（有向グラフ）。
        nmol: 単位胞内のノード数。

    Returns:
        拡大単位胞全体での固定エッジを表す有向グラフ。
    """
    logger = getLogger("replicate_fixed_edges")
    rep_fixed_edges = nx.DiGraph()
    for repi, repj in repgraph.edges():
        i = repi % nmol
        j = repj % nmol
        if fixed.has_edge(i, j):
            rep_fixed_edges.add_edge(repi, repj)
        elif fixed.has_edge(j, i):
            rep_fixed_edges.add_edge(repj, repi)
    for edge in rep_fixed_edges.edges():
        logger.debug(f"* {edge=}")
    return rep_fixed_edges


def replicate_subgraph(
    repgraph: nx.Graph, subgraph: nx.Graph, nmol: int
) -> Generator[nx.Graph, None, None]:
    """単位胞内の subgraph の各レプリカを repgraph から取り出して yield する。

    Args:
        repgraph: 拡大単位胞全体の無向グラフ。
        subgraph: 単位胞内の部分グラフ（例: 1ケージを構成するノードと辺）。
        nmol: 単位胞内のノード数。

    Yields:
        拡大単位胞内の各レプリカに対応する部分グラフ（nx.Graph）。
    """
    origin = list(subgraph.nodes())[0]
    nrep = len(repgraph) // nmol  # number of replicas

    def _next(_reporigin):
        for _repnei in nx.neighbors(repgraph, _reporigin):
            _origin = _reporigin % nmol
            _nei = _repnei % nmol
            if _nei in nx.neighbors(subgraph, _origin) and not replica.has_edge(
                _reporigin, _repnei
            ):
                replica.add_edge(_reporigin, _repnei)
                _next(_repnei)

    for rep in range(nrep):
        reporigin = origin + nmol * rep
        replica = nx.Graph()
        _next(reporigin)
        yield replica


# ============================================================================
# DependencyEngineタスク関数定義: 依存関係は引数名から自動推論される
# 各関数には @reactive を付ける。関数名がそのまま genice.<名前> になるので名詞で書く。
# ============================================================================

_genice3_logger = getLogger("GenIce3")


@reactive
def cell(unitcell: UnitCell, replication_matrix: np.ndarray) -> np.ndarray:
    """拡大単位胞のセル行列"""
    return unitcell.cell @ replication_matrix


@reactive
def replica_vectors(replication_matrix: np.ndarray) -> np.ndarray:
    """レプリカベクトルを計算する。

    拡大単位胞を構成するために必要な、元の単位胞のグリッド位置を表す
    整数ベクトルのリストを生成します。各ベクトルは、拡大単位胞内での
    単位胞の相対位置を表します。

    Args:
        replication_matrix: 単位胞を複製するための3x3整数行列

    Returns:
        np.ndarray: レプリカベクトルの配列（Nx3配列、Nは拡大単位胞内の単位胞数）
    """
    i, j, k = np.array(replication_matrix)
    corners = np.array(
        [a * i + b * j + c * k for a in (0, 1) for b in (0, 1) for c in (0, 1)]
    )

    mins = np.min(corners, axis=0)
    maxs = np.max(corners, axis=0)
    _genice3_logger.debug(f"  {mins=}, {maxs=}")

    det = abs(np.linalg.det(replication_matrix))
    det = np.floor(det + 0.5).astype(int)
    invdet = np.floor(np.linalg.inv(replication_matrix) * det + 0.5).astype(int)

    vecs = set()
    for a in range(mins[0], maxs[0] + 1):
        for b in range(mins[1], maxs[1] + 1):
            for c in range(mins[2], maxs[2] + 1):
                abc = np.array([a, b, c])
                rep = grandcell_wrap(abc, replication_matrix, invdet, det).astype(int)
                if tuple(rep) not in vecs:
                    vecs.add(tuple(rep))

    vecs = np.array(list(vecs))
    vol = abs(np.linalg.det(replication_matrix))
    if not np.allclose(vol, len(vecs)):
        raise ValueError(
            f"replication_matrix determinant ({vol}) must equal number of "
            f"replica vectors ({len(vecs)})"
        )
    return vecs


@reactive
def replica_vector_labels(replica_vectors: np.ndarray) -> Dict[Tuple[int, ...], int]:
    """レプリカベクトル座標タプル → 一意のインデックス の辞書を返す。

    拡大単位胞内での単位胞の位置を、グラフ複製などで参照するために使う。

    Args:
        replica_vectors: レプリカベクトルの配列（Nx3）。

    Returns:
        座標タプル (a, b, c) を 0..N-1 のインデックスにマッピングする辞書。
    """
    return {tuple(xyz): i for i, xyz in enumerate(replica_vectors)}


@reactive
def graph(
    unitcell: UnitCell,
    replica_vectors: np.ndarray,
    replica_vector_labels: Dict[Tuple[int, ...], int],
    replication_matrix: np.ndarray,
) -> nx.Graph:
    """拡大単位胞に対応するグラフを生成する。

    基本単位胞のグラフ（水分子間の水素結合ネットワーク）を、
    拡大単位胞全体に複製して統合したグラフを生成します。
    このグラフは、拡大単位胞内のすべての水分子間の接続関係を表します。

    Args:
        unitcell: 基本単位胞オブジェクト
        replica_vectors: レプリカベクトルの配列
        replica_vector_labels: レプリカベクトル座標タプル→インデックスの辞書（replica_vector_labels タスクの戻り値）
        replication_matrix: 単位胞を複製するための3x3整数行列

    Returns:
        nx.Graph: 拡大単位胞全体の水素結合ネットワークを表す無向グラフ
    """
    g = _replicate_graph(
        unitcell.graph,
        unitcell.lattice_sites,
        replica_vectors,
        replica_vector_index=replica_vector_labels,
        reshape=replication_matrix,
    )
    getLogger(__name__).info("graph: %s", _graph_degree_stats(g))
    return g


@reactive
def lattice_sites(
    unitcell: UnitCell,
    replica_vectors: np.ndarray,
    replication_matrix: np.ndarray,
) -> np.ndarray:
    """格子サイト位置を拡大単位胞全体に複製する。

    基本単位胞内の水分子の座標を、拡大単位胞全体に複製します。
    各水分子の位置は、単位胞の周期境界条件に従って配置されます。

    Args:
        unitcell: 基本単位胞オブジェクト
        replica_vectors: レプリカベクトルの配列
        replication_matrix: 単位胞を複製するための3x3整数行列

    Returns:
        np.ndarray: 拡大単位胞内のすべての水分子の座標（Nx3配列、Nは水分子数）
    """
    return replicate_positions(
        unitcell.lattice_sites, replica_vectors, replication_matrix
    )


@reactive
def anions(
    unitcell: UnitCell, replica_vectors: np.ndarray, spot_anions: Dict[int, str]
) -> Dict[int, str]:
    """アニオンイオンの配置情報を拡大単位胞全体に複製する。

    基本単位胞内で定義されたアニオンイオンと、spot_anionsで指定された
    特定位置のアニオンを統合し、拡大単位胞全体でのアニオン配置を返します。

    Args:
        unitcell: 基本単位胞オブジェクト
        replica_vectors: レプリカベクトルの配列
        spot_anions: 特定の格子サイト位置に配置するアニオンの辞書（サイトインデックス -> イオン名）

    Returns:
        Dict[int, str]: 拡大単位胞全体でのアニオン配置（サイトインデックス -> イオン名）
    """
    anion_dict: Dict[int, str] = {}
    Z = len(unitcell.lattice_sites)
    for site_index, ion_name in unitcell.anions.items():
        for i in range(len(replica_vectors)):
            site = i * Z + site_index
            anion_dict[site] = ion_name
    for site_index, ion_name in spot_anions.items():
        anion_dict[site_index] = ion_name
    return anion_dict


@reactive
def cations(
    unitcell: UnitCell, replica_vectors: np.ndarray, spot_cations: Dict[int, str]
) -> Dict[int, str]:
    """カチオンイオンの配置情報を拡大単位胞全体に複製する。

    基本単位胞内で定義されたカチオンイオンと、spot_cationsで指定された
    特定位置のカチオンを統合し、拡大単位胞全体でのカチオン配置を返します。

    Args:
        unitcell: 基本単位胞オブジェクト
        replica_vectors: レプリカベクトルの配列
        spot_cations: 特定の格子サイト位置に配置するカチオンの辞書（サイトインデックス -> イオン名）

    Returns:
        Dict[int, str]: 拡大単位胞全体でのカチオン配置（サイトインデックス -> イオン名）
    """
    cation_dict: Dict[int, str] = {}
    Z = len(unitcell.lattice_sites)
    for site_index, ion_name in unitcell.cations.items():
        for i in range(len(replica_vectors)):
            site = i * Z + site_index
            cation_dict[site] = ion_name
    for site_index, ion_name in spot_cations.items():
        cation_dict[site_index] = ion_name
    return cation_dict


@reactive
def site_occupants(
    anions: Dict[int, str],
    cations: Dict[int, str],
    spot_hydroniums: List[int],
    spot_hydroxides: List[int],
    lattice_sites: np.ndarray,
) -> List[str]:
    """各格子サイトの占有種（水分子・イオン・H3O+・OH-）のリストを生成する。

    各格子サイトが水分子、アニオン、カチオン、H3O+、OH- のいずれで占有されているかを
    判定し、サイトインデックス順にリストとして返します。

    Args:
        anions: アニオン配置の辞書
        cations: カチオン配置の辞書
        spot_hydroniums: H3O+ を置くサイトのリスト
        spot_hydroxides: OH- を置くサイトのリスト
        lattice_sites: 格子サイト位置の配列

    Returns:
        List[str]: 各サイトの占有種のリスト（"water" またはイオン名）
    """
    occupants = ["water"] * len(lattice_sites)
    for site, ion_name in anions.items():
        occupants[site] = ion_name
    for site, ion_name in cations.items():
        occupants[site] = ion_name
    for site in spot_hydroniums:
        occupants[site] = "H3O+"
    for site in spot_hydroxides:
        occupants[site] = "OH-"
    return occupants


@reactive
def fixed_edges(
    graph: nx.Graph,
    unitcell: UnitCell,
    spot_anions: Dict[int, str],
    spot_cations: Dict[int, str],
    spot_hydroniums: List[int],
    spot_hydroxides: List[int],
    bjerrum_L_edges: List[Tuple[int, int]],
    bjerrum_D_edges: List[Tuple[int, int]],
) -> nx.DiGraph:
    """固定エッジ（水素結合の方向が固定されたエッジ）を拡大単位胞全体に複製する。

    基本単位胞で定義された固定エッジと、spot_anions/spot_cations/spot_hydroniums/
    spot_hydroxides で指定されたイオン・置換種に基づく固定エッジを統合し、
    拡大単位胞全体での固定エッジを返します。
    - アニオン: 4本受け入れ。カチオン: 4本供与。
    - H3O+ (spot_hydroniums): 1本受け入れ・3本供与。
    - OH- (spot_hydroxides): 3本受け入れ・1本供与。

    Args:
        graph: 拡大単位胞全体のグラフ
        unitcell: 基本単位胞オブジェクト
        spot_anions: 特定位置のアニオン配置
        spot_cations: 特定位置のカチオン配置
        spot_hydroniums: H3O+ を置くサイトのリスト（1受容・3供与）
        spot_hydroxides: OH- を置くサイトのリスト（3受容・1供与）

    Returns:
        nx.DiGraph: 拡大単位胞全体での固定エッジを表す有向グラフ
    """
    if (
        spot_anions
        or spot_cations
        or spot_hydroniums
        or spot_hydroxides
        or bjerrum_L_edges
        or bjerrum_D_edges
    ) and not unitcell.SUPPORTS_ION_DOPING:
        raise ConfigurationError(
            "Ion doping (spot_anion/spot_cation/spot_hydronium/spot_hydroxide) "
            "is not supported for hydrogen-ordered ices."
        )
    dg = _replicate_fixed_edges(graph, unitcell.fixed, len(unitcell.lattice_sites))
    for site_index in spot_anions:
        for nei in graph.neighbors(site_index):
            if dg.has_edge(site_index, nei):
                raise ConfigurationError(
                    f"矛盾する辺の固定 at {site_index}; すでに({site_index} --> {nei})が固定されています。"
                )
            else:
                dg.add_edge(nei, site_index)
    for site_index in spot_cations:
        for nei in graph.neighbors(site_index):
            if dg.has_edge(nei, site_index):
                raise ConfigurationError(
                    f"矛盾する辺の固定 at {site_index}; すでに({nei} --> {site_index})が固定されています。"
                )
            else:
                dg.add_edge(site_index, nei)
    for site_index in spot_hydroniums:
        neis = sorted(graph.neighbors(site_index))
        if len(neis) != 4:
            raise ConfigurationError(
                f"spot_hydronium at {site_index} must have 4 neighbors, got {len(neis)}."
            )
        dg.add_edge(neis[0], site_index)
        for nei in neis[1:]:
            dg.add_edge(site_index, nei)
    for site_index in spot_hydroxides:
        neis = sorted(graph.neighbors(site_index))
        if len(neis) != 4:
            raise ConfigurationError(
                f"spot_hydroxide at {site_index} must have 4 neighbors, got {len(neis)}."
            )
        dg.add_edge(site_index, neis[0])
        for nei in neis[1:]:
            dg.add_edge(nei, site_index)
    # Bjerrum L 欠陥: i ノードを hydronium と同様に扱う。
    for i, j in bjerrum_L_edges or []:
        neis = sorted(graph.neighbors(i))
        if len(neis) != 4:
            raise ConfigurationError(
                f"bjerrum_L at {i} must have 4 neighbors, got {len(neis)}."
            )
        if j not in neis:
            raise ConfigurationError(
                f"bjerrum_L edge ({i}, {j}) is not an edge in the graph."
            )
        # 受容側となる隣接ノードを j 以外から 1 つ選ぶ
        acceptor = next((nei for nei in neis if nei != j), None)
        if acceptor is None:
            raise ConfigurationError(
                f"bjerrum_L at {i} cannot choose acceptor neighbor distinct from {j}."
            )
        # hydronium と同様のチェック・追加ロジック
        for nei in neis:
            if nei == acceptor:
                # nei -> i （受容）
                if dg.has_edge(i, nei):
                    raise ConfigurationError(
                        f"矛盾する辺の固定 at {i}; すでに({i} --> {nei})が固定されています。"
                    )
                dg.add_edge(nei, i)
            else:
                # i -> nei （供与）
                if dg.has_edge(nei, i):
                    raise ConfigurationError(
                        f"矛盾する辺の固定 at {i}; すでに({nei} --> {i})が固定されています。"
                    )
                dg.add_edge(i, nei)
    # Bjerrum D 欠陥: i ノードを hydroxide と同様に扱う（ただし j は供与側にしない）。
    for i, j in bjerrum_D_edges or []:
        neis = sorted(graph.neighbors(i))
        if len(neis) != 4:
            raise ConfigurationError(
                f"bjerrum_D at {i} must have 4 neighbors, got {len(neis)}."
            )
        if j not in neis:
            raise ConfigurationError(
                f"bjerrum_D edge ({i}, {j}) is not an edge in the graph."
            )
        # 供与側となる隣接ノードを j 以外から 1 つ選ぶ
        donor = next((nei for nei in neis if nei != j), None)
        if donor is None:
            raise ConfigurationError(
                f"bjerrum_D at {i} cannot choose donor neighbor distinct from {j}."
            )
        for nei in neis:
            if nei == donor:
                # i -> nei （供与）
                if dg.has_edge(nei, i):
                    raise ConfigurationError(
                        f"矛盾する辺の固定 at {i}; すでに({nei} --> {i})が固定されています。"
                    )
                dg.add_edge(i, nei)
            else:
                # nei -> i （受容）
                if dg.has_edge(i, nei):
                    raise ConfigurationError(
                        f"矛盾する辺の固定 at {i}; すでに({i} --> {nei})が固定されています。"
                    )
                dg.add_edge(nei, i)
    return dg


@reactive
def digraph(
    graph: nx.Graph,
    depol_loop: int,
    lattice_sites: np.ndarray,
    fixed_edges: nx.DiGraph,
    target_pol: np.ndarray,
    bjerrum_L_edges: List[Tuple[int, int]],
    bjerrum_D_edges: List[Tuple[int, int]],
) -> nx.DiGraph:
    """水素結合ネットワークの有向グラフを生成する。

    無向グラフ（水素結合ネットワーク）から、各水素結合の方向（プロトンの向き）
    を決定して有向グラフを生成します。固定エッジで指定された方向は維持され、
    それ以外のエッジは双極子最適化アルゴリズムにより方向が決定されます。

    Args:
        graph: 拡大単位胞全体の無向グラフ
        depol_loop: 双極子最適化の反復回数
        lattice_sites: 格子サイト位置の配列
        fixed_edges: 拡大単位胞全体での固定エッジの有向グラフ
        target_pol: 分極の目標値

    Returns:
        nx.DiGraph: 水素結合の方向が決定された有向グラフ
    """
    for edge in fixed_edges.edges():
        _genice3_logger.debug(f"+ {edge=}")
    dg = genice_core.ice_graph(
        graph,
        vertex_positions=lattice_sites,
        is_periodic_boundary=True,
        dipole_optimization_cycles=depol_loop,
        fixed_edges=fixed_edges,
        pairing_attempts=1000,
        target_pol=target_pol,
    )
    if not dg:
        raise ConfigurationError("Failed to generate a directed graph.")
    # Bjerrum L 欠陥: 対応するエッジを削除する（両方向とも取り除く）
    for edge in bjerrum_L_edges or []:
        i, j = edge
        if dg.has_edge(i, j):
            dg.remove_edge(i, j)
        if dg.has_edge(j, i):
            dg.remove_edge(j, i)
    # Bjerrum D 欠陥: 対応するエッジを両方向とも存在させる
    for edge in bjerrum_D_edges or []:
        i, j = edge
        if not dg.has_edge(i, j):
            dg.add_edge(i, j)
        if not dg.has_edge(j, i):
            dg.add_edge(j, i)
    return dg


@reactive
def orientations(
    lattice_sites: np.ndarray,
    digraph: nx.DiGraph,
    cell: np.ndarray,
    anions: Dict[int, str],
    cations: Dict[int, str],
    spot_hydroniums: List[int],
    spot_hydroxides: List[int],
) -> np.ndarray:
    """各水分子の配向行列を計算する。

    有向グラフで決定された水素結合の方向に基づいて、各水分子の
    配向（回転行列）を計算します。水分子のOHベクトルの方向から
    分子全体の配向を決定します。

    Args:
        lattice_sites: 格子サイト位置の配列
        digraph: 水素結合の方向が決定された有向グラフ
        cell: 拡大単位胞のセル行列
        anions: アニオン配置の辞書
        cations: カチオン配置の辞書
        spot_hydroniums: H3O+ を置くサイトのリスト
        spot_hydroxides: OH- を置くサイトのリスト

    Returns:
        np.ndarray: 各水分子の配向行列（Nx3x3配列、Nは水分子数）
    """
    return _assume_water_orientations(
        lattice_sites,
        digraph,
        cell,
        set(anions) | set(cations) | set(spot_hydroniums) | set(spot_hydroxides),
    )


@reactive
def cages(
    unitcell: UnitCell,
    replica_vectors: np.ndarray,
    replication_matrix: np.ndarray,
    graph: nx.Graph,
) -> CageSpecs | None:
    """ケージ位置とタイプを拡大単位胞全体に複製する。

    基本単位胞内で定義されたゲスト分子を配置するためのケージ（空隙）の
    位置とタイプを、拡大単位胞全体に複製する。

    Args:
        unitcell: 基本単位胞オブジェクト。
        replica_vectors: レプリカベクトルの配列。
        replication_matrix: 単位胞を複製するための3x3整数行列。
        graph: 拡大単位胞全体の無向グラフ（サブグラフ複製に使用）。

    Returns:
        CageSpecs: 拡大単位胞全体でのケージ位置とタイプ。
        unitcell.cages が None の場合は None。
    """
    if unitcell.cages is None:
        return None
    # ケージが存在しない場合（空の配列）の処理
    if len(unitcell.cages.positions) == 0:
        return CageSpecs(positions=np.array([]).reshape(0, 3), specs=[])
    repcagepos = replicate_positions(
        unitcell.cages.positions, replica_vectors, replication_matrix
    )
    num_cages_in_unitcell = len(unitcell.cages.positions)

    # replicate_subgraphはケージ一個ずつをreplicateするが、repcagespecsの並び順は単位胞単位。
    repcagespecs = [None] * len(repcagepos)
    for i, cage in enumerate(unitcell.cages.specs):
        for j, repgraph in enumerate(
            replicate_subgraph(graph, cage.graph, unitcell.lattice_sites.shape[0])
        ):
            repcagespecs[i + j * num_cages_in_unitcell] = CageSpec(
                cage_type=cage.cage_type, faces=cage.faces, graph=repgraph
            )
    # unit cellのcagesと同じ構造。
    _genice3_logger.debug(f"{repcagepos=}, {repcagespecs=}")
    return CageSpecs(positions=repcagepos, specs=repcagespecs)


def place_groups_on_lattice(
    genice: "GenIce3",
    spot_cation_groups: Dict[int, Dict[int, str]],
) -> None:
    """
    格子点に group を配置する。

    Args:
        genice: GenIce3 インスタンス
        spot_cation_groups: サイト -> {ケージID -> group名} の辞書
    """
    _genice3_logger.info(
        f"place_groups_on_lattice: dummy implementation (group assignment not yet applied) {spot_cation_groups=}"
    )


def log_spot_cation_cages(genice: "GenIce3") -> None:
    """spot_cation ごとに属するケージのID・cage_type・facesをログ表示する。"""
    if not genice.spot_cations or genice.cages is None or len(genice.cages.specs) == 0:
        return
    num_replicas = int(np.round(np.linalg.det(genice.replication_matrix)))
    num_cages_in_unitcell = len(genice.cages.specs) // max(1, num_replicas)
    for site, ion_name in genice.spot_cations.items():
        cage_indices = genice.cages.node_to_cage_indices.get(site, [])
        for cage_id in cage_indices:
            spec = genice.cages.specs[cage_id]
            genice.logger.info(
                f"spot_cation {site}={ion_name} belongs to cage {cage_id} ({spec.cage_type} {spec.faces}) in the supercell."
            )


def log_cation_cages(genice: "GenIce3") -> None:
    """単位胞内のカチオンごとに、属するケージのID・cage_type・faces をログ表示する。"""
    if (
        not genice.cations
        or genice.unitcell.cages is None
        or len(genice.unitcell.cages.specs) == 0
    ):
        return
    for site, ion_name in genice.cations.items():
        cage_indices = genice.unitcell.cages.node_to_cage_indices.get(site, [])
        for cage_id in cage_indices:
            spec = genice.unitcell.cages.specs[cage_id]
            genice.logger.info(
                f"cation {site}={ion_name} belongs to cage {cage_id} ({spec.cage_type} {spec.faces}) in the unit cell."
            )


def place_group(direction: np.ndarray, bondlen: float, group_name: str) -> Group:
    """指定方向・結合長で group プラグインを読み込み、配置した Group を返す。

    置換イオンから group のアンカー原子までが direction 方向に bondlen の長さで並ぶ。
    TODO: 将来は2個以上のアンカーを持つ group を扱う可能性がある。

    Args:
        direction: 配置方向（直交座標、正規化される）。
        bondlen: アンカーまでの結合長。
        group_name: group プラグイン名（例: "ammonia"）。

    Returns:
        回転・並進を適用した Group インスタンス。
    """
    # logger = getLogger("GenIce3")
    group = safe_import("group", group_name).Group()
    # logger.info(f"{group_name=} {group=}")
    direction /= np.linalg.norm(direction)
    offset = direction * bondlen
    # Z軸をdirection方向に傾ける行列
    ex, ey, ez = direction
    r = (ex**2 + ey**2) ** 0.5
    Ry = np.array([[ez, 0, -r], [0, 1, 0], [r, 0, ez]])
    Rz = np.array([[ex / r, ey / r, 0], [-ey / r, ex / r, 0], [0, 0, 1]])
    rotmat = Ry @ Rz
    sites = group.sites @ rotmat + offset
    return Group(
        sites=sites,
        labels=group.labels,
        bonds=group.bonds,
        name=group.name,
    )


# ============================================================================
# GenIce3クラス: DependencyEngineをラップ
# ============================================================================


class GenIce3:
    """GenIce3のメインクラス：リアクティブプロパティによる氷構造生成システム

    GenIce3は、依存関係エンジン（DependencyEngine）を使用して、必要な時に
    自動的に計算されるリアクティブプロパティを提供します。これにより、ユーザーは
    必要なプロパティにアクセスするだけで、そのプロパティに依存するすべての
    計算が自動的に実行されます。

    リアクティブプロパティの仕組み:
        - 各プロパティ（digraph, graph, lattice_sitesなど）は、アクセス時に
          必要に応じて自動的に計算されます
        - 依存関係は関数の引数名から自動的に推論されます
        - 一度計算されたプロパティはキャッシュされ、依存する入力が変更されるまで
          再利用されます
        - 入力プロパティ（unitcell, replication_matrix, depol_loopなど）が
          変更されると、それに依存するすべてのプロパティのキャッシュが自動的に
          クリアされます

    使用例:
        >>> genice = GenIce3(unitcell=my_unitcell)
        >>> digraph = genice.digraph  # 自動的に必要な計算が実行される
        >>> orientations = genice.orientations  # digraphに依存するため、digraphが先に計算される

    Attributes:
        digraph (nx.DiGraph): 水素結合の方向が決定された有向グラフ。
            無向グラフから双極子最適化アルゴリズムにより各水素結合の方向を決定した
            有向グラフです。固定エッジで指定された方向は維持され、それ以外のエッジは
            最適化により決定されます。このプロパティにアクセスすると、必要な依存関係
            （graph, lattice_sites, fixed_edges など）が自動的に計算されます。

        graph (nx.Graph): 水素結合ネットワークの無向グラフ。
            拡大単位胞全体の水分子間の水素結合ネットワークを表す無向グラフです。
            基本単位胞のグラフを拡大単位胞全体に複製して統合したものです。

        lattice_sites (np.ndarray): 格子サイト位置の配列。
            拡大単位胞内のすべての水分子の座標を表すNx3配列です（Nは水分子数）。
            基本単位胞内の水分子の座標を、単位胞の周期境界条件に従って拡大単位胞全体に
            複製したものです。

        orientations (np.ndarray): 各水分子の配向行列。
            各水分子の配向（回転行列）を表すNx3x3配列です（Nは水分子数）。
            有向グラフで決定された水素結合の方向に基づいて、各水分子のOHベクトルの
            方向から分子全体の配向を決定します。

        unitcell (UnitCell): 基本単位胞オブジェクト。
            氷構造の基本単位胞を表すオブジェクトです。格子構造、水分子の配置、
            水素結合ネットワーク、固定エッジなどの情報を含みます。

        replication_matrix (np.ndarray): 単位胞複製行列。
            基本単位胞をどのように積み重ねて拡大単位胞を作成するかを指定する3x3整数行列です。
            単位行列の場合は基本単位胞のみを使用します。

        depol_loop (int): 双極子最適化の反復回数。
            有向グラフを生成する際の双極子最適化アルゴリズムの反復回数です。
            値が大きいほどより最適化された構造が得られますが、計算時間も増加します。

        seed (int): 乱数シード。
            乱数生成器のシード値です。digraphの生成などで使用される乱数の初期化に使用されます。
            このプロパティを変更すると、それに依存するすべてのリアクティブプロパティの
            キャッシュが自動的にクリアされます。

        spot_anions (Dict[int, str]): 特定の格子サイト位置に配置するアニオンイオンの辞書。
            サイトインデックスからイオン名へのマッピングです。このプロパティを変更すると、
            それに依存するすべてのリアクティブプロパティのキャッシュが自動的にクリアされます。

        spot_cations (Dict[int, str]): 特定の格子サイト位置に配置するカチオンイオンの辞書。
            サイトインデックスからイオン名へのマッピングです。このプロパティを変更すると、
            それに依存するすべてのリアクティブプロパティのキャッシュが自動的にクリアされます。

        spot_hydroniums (List[int]): H3O+ を置くサイトのリスト（1受容・3供与）。

        spot_hydroxides (List[int]): OH- を置くサイトのリスト（3受容・1供与）。

        cages (CageSpecs): 拡大単位胞全体でのケージ位置・タイプ。
            単位胞のケージを replica_vectors に従って複製したもので、ゲスト配置や cage_survey 出力に利用します。

        fixed_edges (nx.DiGraph): 拡大単位胞全体での固定エッジの有向グラフ。
            単位胞の固定エッジと spot_anion/spot_cation/spot_hydronium/spot_hydroxide に基づく固定を統合したもので、digraph の生成に利用します。
    """

    # Class名でlog表示したい。
    logger = getLogger("GenIce3")

    # ユーザー向けAPIとして公開するpropertyのリスト
    # このリストに含まれるpropertyのみAPIドキュメントを作成する
    PUBLIC_API_PROPERTIES = [
        "digraph",
        "graph",
        "lattice_sites",
        "orientations",
        "unitcell",
        "replication_matrix",
        "depol_loop",
        "target_pol",
        "seed",
        "spot_anions",
        "spot_cations",
        "spot_hydroniums",
        "spot_hydroxides",
    ]

    def __init__(
        self,
        depol_loop: int = 1000,
        replication_matrix: np.ndarray = np.eye(3, dtype=int),
        target_pol: np.ndarray = np.array([0.0, 0.0, 0.0]),
        seed: int = 1,
        spot_anions: Dict[int, str] = {},
        spot_cations: Dict[int, str] = {},
        spot_hydroniums: List[int] = None,
        spot_hydroxides: List[int] = None,
        bjerrum_L_edges: List[Tuple[int, int]] | None = None,
        bjerrum_D_edges: List[Tuple[int, int]] | None = None,
        guests: Dict[str, List["GuestSpec"]] = None,
        spot_guests: Dict[int, Molecule] = None,
        spot_cation_groups: Dict[int, Dict[int, str]] = None,
        **kwargs: Any,
    ) -> None:
        """GenIce3インスタンスを初期化する。

        Args:
            depol_loop: 双極子最適化の反復回数（デフォルト: 1000）
            replication_matrix: 単位胞複製行列（デフォルト: 単位行列）
            target_pol: 分極の目標値（デフォルト: [0, 0, 0]）
            seed: 乱数シード（デフォルト: 1）
            spot_anions: 特定位置のアニオン配置（デフォルト: {}）
            spot_cations: 特定位置のカチオン配置（デフォルト: {}）
            spot_hydroniums: H3O+ を置くサイトのリスト（デフォルト: []）
            spot_hydroxides: OH- を置くサイトのリスト（デフォルト: []）
            guests: ケージタイプごとのゲスト分子指定（デフォルト: {}）
            spot_guests: 特定ケージ位置へのゲスト分子指定（デフォルト: {}）
            spot_cation_groups: spot_cation の --group 指定（サイト -> {ケージID -> group名}）（デフォルト: {}）
            **kwargs: その他のリアクティブプロパティ（unitcellなど）

        Raises:
            ConfigurationError: 無効なキーワード引数が指定された場合
        """
        # DependencyEngineインスタンスを作成
        self.engine = DependencyEngine()

        # Default値が必要なもの
        self.seed = (
            seed  # reactive propertyとして設定（setterでnp.random.seed()も実行される）
        )
        self.depol_loop = depol_loop
        self.replication_matrix = replication_matrix
        self.target_pol = target_pol
        self.spot_anions = spot_anions
        self.spot_cations = spot_cations
        self.spot_hydroniums = spot_hydroniums if spot_hydroniums is not None else []
        self.spot_hydroxides = spot_hydroxides if spot_hydroxides is not None else []
        self.bjerrum_L_edges = bjerrum_L_edges if bjerrum_L_edges is not None else []
        self.bjerrum_D_edges = bjerrum_D_edges if bjerrum_D_edges is not None else []
        self.guests = guests if guests is not None else {}
        self.spot_guests = spot_guests if spot_guests is not None else {}
        self.spot_cation_groups = (
            spot_cation_groups if spot_cation_groups is not None else {}
        )

        # タスクを登録（モジュールレベルの関数を登録）
        self._register_tasks()

        # Default値が不要なもの
        for key in self.list_settable_reactive_properties():
            if key in kwargs:
                setattr(self, key, kwargs.pop(key))
        if kwargs:
            raise ConfigurationError(f"Invalid keyword arguments: {kwargs}.")

    def _register_tasks(self):
        """DependencyEngine に @reactive を付けたタスク関数を登録する"""
        for func in get_reactive_tasks(__name__):
            self.engine.task(func)

    # spot_anions
    @property
    def spot_anions(self):
        """特定の格子サイト位置に配置するアニオンイオンの辞書。

        Returns:
            Dict[int, str]: サイトインデックスからイオン名へのマッピング
        """
        if not hasattr(self, "_spot_anions"):
            self._spot_anions = {}
        return self._spot_anions

    @spot_anions.setter
    def spot_anions(self, spot_anions):
        """特定の格子サイト位置に配置するアニオンイオンを設定する。

        このプロパティを変更すると、それに依存するすべてのリアクティブプロパティの
        キャッシュが自動的にクリアされます。

        Args:
            spot_anions: サイトインデックスからイオン名へのマッピング
        """
        self._spot_anions = spot_anions
        self.logger.debug(f"  {spot_anions=}")
        self.engine.cache.clear()

    # spot_cations
    @property
    def spot_cations(self):
        """特定の格子サイト位置に配置するカチオンイオンの辞書。

        Returns:
            Dict[int, str]: サイトインデックスからイオン名へのマッピング
        """
        if not hasattr(self, "_spot_cations"):
            self._spot_cations = {}
        return self._spot_cations

    @spot_cations.setter
    def spot_cations(self, spot_cations):
        """特定の格子サイト位置に配置するカチオンイオンを設定する。

        このプロパティを変更すると、それに依存するすべてのリアクティブプロパティの
        キャッシュが自動的にクリアされます。

        Args:
            spot_cations: サイトインデックスからイオン名へのマッピング
        """
        self._spot_cations = spot_cations
        self.logger.debug(f"  {spot_cations=}")
        self.engine.cache.clear()

    # spot_hydroniums (H3O+: 1受容・3供与)
    @property
    def spot_hydroniums(self):
        """H3O+ を置くサイトのリスト。1本受け入れ・3本供与。

        Returns:
            List[int]: サイトインデックスのリスト
        """
        if not hasattr(self, "_spot_hydroniums"):
            self._spot_hydroniums = []
        return self._spot_hydroniums

    @spot_hydroniums.setter
    def spot_hydroniums(self, spot_hydroniums):
        self._spot_hydroniums = (
            list(spot_hydroniums) if spot_hydroniums is not None else []
        )
        self.logger.debug(f"  {spot_hydroniums=}")
        self.engine.cache.clear()

    # spot_hydroxides (OH-: 3受容・1供与)
    @property
    def spot_hydroxides(self):
        """OH- を置くサイトのリスト。3本受け入れ・1本供与。

        Returns:
            List[int]: サイトインデックスのリスト
        """
        if not hasattr(self, "_spot_hydroxides"):
            self._spot_hydroxides = []
        return self._spot_hydroxides

    @spot_hydroxides.setter
    def spot_hydroxides(self, spot_hydroxides):
        self._spot_hydroxides = (
            list(spot_hydroxides) if spot_hydroxides is not None else []
        )
        self.logger.debug(f"  {spot_hydroxides=}")
        self.engine.cache.clear()

    # Bjerrum L 欠陥に対応するエッジ集合（(i, j) のリスト）
    @property
    def bjerrum_L_edges(self) -> List[Tuple[int, int]]:
        if not hasattr(self, "_bjerrum_L_edges"):
            self._bjerrum_L_edges = []
        return self._bjerrum_L_edges

    @bjerrum_L_edges.setter
    def bjerrum_L_edges(self, edges: List[Tuple[int, int]] | None):
        self._bjerrum_L_edges = list(edges) if edges is not None else []
        self.logger.debug(f"  {self._bjerrum_L_edges=}")
        self.engine.cache.clear()

    # Bjerrum D 欠陥に対応するエッジ集合（(i, j) のリスト）
    @property
    def bjerrum_D_edges(self) -> List[Tuple[int, int]]:
        if not hasattr(self, "_bjerrum_D_edges"):
            self._bjerrum_D_edges = []
        return self._bjerrum_D_edges

    @bjerrum_D_edges.setter
    def bjerrum_D_edges(self, edges: List[Tuple[int, int]] | None):
        self._bjerrum_D_edges = list(edges) if edges is not None else []
        self.logger.debug(f"  {self._bjerrum_D_edges=}")
        self.engine.cache.clear()

    def add_spot_hydronium(self, sites):
        """H3O+ を置くサイトを追加登録するヘルパー。

        Args:
            sites: サイトインデックス（int）またはその反復可能オブジェクト。
        """
        if sites is None:
            return
        # numpy のスカラーも受け付ける
        if isinstance(sites, (int, np.integer)):
            new_sites = [int(sites)]
        else:
            new_sites = list(sites)
        # setter を経由してキャッシュを無効化する
        self.spot_hydroniums = list(self.spot_hydroniums) + new_sites

    def add_spot_hydroxide(self, sites):
        """OH- を置くサイトを追加登録するヘルパー。

        Args:
            sites: サイトインデックス（int）またはその反復可能オブジェクト。
        """
        if sites is None:
            return
        if isinstance(sites, (int, np.integer)):
            new_sites = [int(sites)]
        else:
            new_sites = list(sites)
        self.spot_hydroxides = list(self.spot_hydroxides) + new_sites

    def add_bjerrum_L(self, edges):
        """Bjerrum L 欠陥を追加登録するヘルパー。

        Args:
            edges: (i, j) のタプル、またはその反復可能オブジェクト。
        """
        if edges is None:
            return
        # 単一タプル (i, j) も許容する
        if isinstance(edges, tuple) and len(edges) == 2:
            new_edges = [edges]
        else:
            new_edges = list(edges)
        self.bjerrum_L_edges = list(self.bjerrum_L_edges) + new_edges

    def add_bjerrum_D(self, edges):
        """Bjerrum D 欠陥を追加登録するヘルパー。

        Args:
            edges: (i, j) のタプル、またはその反復可能オブジェクト。
        """
        if edges is None:
            return
        if isinstance(edges, tuple) and len(edges) == 2:
            new_edges = [edges]
        else:
            new_edges = list(edges)
        self.bjerrum_D_edges = list(self.bjerrum_D_edges) + new_edges

    @property
    def seed(self):
        """乱数シード。

        Returns:
            int: 乱数シード値
        """
        return self._seed

    @seed.setter
    def seed(self, seed):
        """乱数シードを設定する。

        このプロパティを変更すると、それに依存するすべてのリアクティブプロパティの
        キャッシュが自動的にクリアされます。

        Args:
            seed: 乱数シード値
        """
        self._seed = seed
        np.random.seed(seed)
        self.logger.debug(f"  {seed=}")
        # キャッシュをクリア（seedに依存するすべてのタスクを再計算させる）
        self.engine.cache.clear()

    @property
    def depol_loop(self):
        """双極子最適化の反復回数。

        Returns:
            int: 双極子最適化アルゴリズムの反復回数
        """
        return self._depol_loop

    @depol_loop.setter
    def depol_loop(self, depol_loop):
        """双極子最適化の反復回数を設定する。

        このプロパティを変更すると、それに依存するすべてのリアクティブプロパティの
        キャッシュが自動的にクリアされます。

        Args:
            depol_loop: 双極子最適化アルゴリズムの反復回数
        """
        self._depol_loop = depol_loop
        self.logger.debug(f"  {depol_loop=}")
        # キャッシュをクリア（depol_loopに依存するすべてのタスクを再計算させる）
        self.engine.cache.clear()

    @property
    def target_pol(self):
        """分極の目標値（3要素のベクトル）。"""
        return self._target_pol

    @target_pol.setter
    def target_pol(self, target_pol):
        """分極の目標値を設定する。変更時は依存するキャッシュをクリアする。"""
        self._target_pol = np.asarray(target_pol, dtype=float).reshape(3)
        self.logger.debug(f"  {target_pol=}")
        self.engine.cache.clear()

    @property
    def unitcell(self):
        """基本単位胞オブジェクト。

        Returns:
            UnitCell: 基本単位胞オブジェクト

        Raises:
            ConfigurationError: 単位胞が設定されていない場合
        """
        if not hasattr(self, "_unitcell") or self._unitcell is None:
            raise ConfigurationError("Unitcell is not set.")
        return self._unitcell

    @unitcell.setter
    def unitcell(self, unitcell):
        """基本単位胞オブジェクトを設定する。

        このプロパティを変更すると、それに依存するすべてのリアクティブプロパティの
        キャッシュが自動的にクリアされます。

        Args:
            unitcell: 基本単位胞オブジェクト
        """
        self._unitcell = unitcell
        self.logger.debug(f"  {unitcell=}")
        self.logger.debug(f"  {unitcell.lattice_sites=}")
        self.logger.debug(f"  {unitcell.graph=}")
        self.logger.debug(f"  {unitcell.fixed=}")

        a, b, c, A, B, C = cellshape(self.replication_matrix @ self.unitcell.cell)
        self.logger.debug("  Reshaped cell:")
        self.logger.debug(f"    {a=:.4f}, {b=:.4f}, {c=:.4f}")
        self.logger.debug(f"    {A=:.3f}, {B=:.3f}, {C=:.3f}")
        # キャッシュをクリア（unitcellに依存するすべてのタスクを再計算させる）
        self.engine.cache.clear()

    def set_unitcell(self, unitcell_or_name, **kwargs):
        """基本単位胞を設定する（リアクティブプロパティ用のラッパー）。

        代入（genice.unitcell = ...）の代わりにこのメソッドを使うと、
        単位胞がリアクティブな設定であることが明確になります。

        Args:
            unitcell_or_name: 基本単位胞オブジェクト、または単位胞名（文字列）。
                文字列の場合は UnitCell(name, **kwargs) として内部で生成します。
            **kwargs: unitcell_or_name が文字列の場合に UnitCell(...) に渡すオプション。
        """
        if isinstance(unitcell_or_name, UnitCell):
            self.unitcell = unitcell_or_name
        else:
            self.unitcell = UnitCellPlugin(unitcell_or_name, **kwargs)

    @property
    def replication_matrix(self):
        """単位胞を複製するための3x3整数行列。

        この行列により、基本単位胞をどのように積み重ねて拡大単位胞を
        作成するかを指定します。

        Returns:
            np.ndarray: 3x3整数行列
        """
        return self._replication_matrix

    @replication_matrix.setter
    def replication_matrix(self, replication_matrix):
        """単位胞複製行列を設定する。

        このプロパティを変更すると、それに依存するすべてのリアクティブプロパティの
        キャッシュが自動的にクリアされます。

        Args:
            replication_matrix: 3x3整数行列
        """
        self._replication_matrix = np.array(replication_matrix).reshape(3, 3)
        i, j, k = self._replication_matrix
        self.logger.debug(f"    {i=}")
        self.logger.debug(f"    {j=}")
        self.logger.debug(f"    {k=}")
        # キャッシュをクリア（replication_matrixに依存するすべてのタスクを再計算させる）
        self.engine.cache.clear()

    def set_replication_matrix(self, replication_matrix):
        """単位胞複製行列を設定する（リアクティブプロパティ用のラッパー）。

        代入（genice.replication_matrix = ...）の代わりにこのメソッドを使うと、
        複製行列がリアクティブな設定であることが明確になります。

        Args:
            replication_matrix: 3x3整数行列（リストのリストまたは ndarray）
        """
        self.replication_matrix = replication_matrix

    def _get_inputs(self) -> Dict[str, Any]:
        """engine.resolve()に渡すinputs辞書を取得"""
        return {
            "unitcell": self.unitcell,
            "replication_matrix": self.replication_matrix,
            "depol_loop": self.depol_loop,
            "target_pol": self.target_pol,
            "spot_anions": self.spot_anions,
            "spot_cations": self.spot_cations,
            "spot_hydroniums": self.spot_hydroniums,
            "spot_hydroxides": self.spot_hydroxides,
            "bjerrum_L_edges": self.bjerrum_L_edges,
            "bjerrum_D_edges": self.bjerrum_D_edges,
        }

    def __getattr__(self, name: str):
        """リアクティブプロパティへのアクセスを自動解決する。

        このメソッドにより、DependencyEngineに登録されたタスク関数が
        プロパティとしてアクセス可能になります。プロパティにアクセスすると、
        依存関係が自動的に解決され、必要な計算が実行されます。

        例:
            >>> genice = GenIce3(unitcell=my_unitcell)
            >>> digraph = genice.digraph  # この時点でdigraphとその依存関係が計算される

        Args:
            name: アクセスするプロパティ名（タスク関数名）

        Returns:
            計算されたプロパティの値

        Raises:
            AttributeError: 指定された名前のプロパティが存在しない場合
            ConfigurationError: 必要な入力が設定されていない場合
        """
        # DependencyEngineに登録されているタスクかチェック
        if name in self.engine.registry:
            return self.engine.resolve(name, self._get_inputs())

        # それ以外は通常のAttributeErrorを発生
        raise AttributeError(
            f"'{self.__class__.__name__}' object has no attribute '{name}'"
        )

    def water_molecules(self, water_model: Molecule) -> Dict[int, Molecule]:
        """格子サイトが水のサイトについて、配向・位置を適用した水分子の辞書を返す。

        Args:
            water_model: 水分子のテンプレート（sites, labels など）。

        Returns:
            サイト番号（ノード番号）→ Molecule の辞書。イオンサイトは含まない。
        """
        mols = {}
        for site in range(len(self.lattice_sites)):
            occ = self.site_occupants[site]
            # hydroniumやhydroxideはwater-likeではない。
            is_water_like = occ == "water"
            if is_water_like:
                rel_position = self.lattice_sites[site]
                orientation = self.orientations[site]

                sites = water_model.sites @ orientation + rel_position @ self.cell
                mols[site] = Molecule(
                    name=water_model.name,
                    sites=sites,
                    labels=water_model.labels,
                    is_water=True,
                )
        return mols

    def guest_molecules(self) -> List[Molecule]:
        """GenIce3に設定されたguestsとspot_guestsからゲスト分子リストを生成する。"""
        # 通常の氷（ゲストなし）ではケージ不要 → assess_cages を呼ばない
        if not self.guests and not self.spot_guests:
            return []

        all_cage_types = [spec.cage_type for spec in self.cages.specs]
        available = sorted(set(all_cage_types))
        # guest_specで指定されているのに存在しない種類のケージはエラーにする。
        for cage_type in self.guests:
            if cage_type not in all_cage_types:
                raise ConfigurationError(
                    f"Cage type {cage_type} is not defined. "
                    f"Available cage types in this structure: {available}."
                )

        randoms = np.random.random(len(self.cages.positions))

        mols = []
        for pos, spec, probability in zip(
            self.cages.positions, self.cages.specs, randoms
        ):
            accum = 0.0
            if spec.cage_type in self.guests:
                for guest_spec in self.guests[spec.cage_type]:
                    molecule = guest_spec.molecule
                    occupancy = guest_spec.occupancy
                    accum += occupancy
                    if accum > probability:
                        mols.append(
                            Molecule(
                                name=molecule.name,
                                sites=molecule.sites + pos @ self.cell,
                                labels=molecule.labels,
                                is_water=molecule.is_water,
                            )
                        )
                        break

        # spot guestの配置
        for cage_index, guest in self.spot_guests.items():
            mols.append(
                Molecule(
                    name=guest.name,
                    sites=guest.sites + self.cages.positions[cage_index] @ self.cell,
                    labels=guest.labels,
                    is_water=False,
                )
            )
        return mols

    def build_molecular_ion(
        self,
        site_index: int,
        molecule: str,
        groups: Dict[int, str] | None = None,
    ) -> Molecule:
        """サイト site_index（格子サイトのノード番号）に分子イオン（および spot cation 用の修飾 group）を構築する。

        Args:
            site_index: 格子サイトのノード番号。
            molecule: 分子イオンのプラグイン名（例: "ammonium"）。
            groups: 当該サイトに隣接するケージID → group名。spot cation の修飾用。

        Returns:
            直交座標で配置された Molecule（原子名は labels）。
        """
        groups = groups or {}
        ion_center = self.lattice_sites[site_index] @ self.cell
        try:
            ion = safe_import("molecule", molecule).Molecule()
            name = ion.name
            sites = ion.sites + ion_center
            atom_names: List[str] = list(ion.labels)
        except ImportError:
            name = molecule
            sites = np.array([ion_center])
            atom_names = [molecule]
        # 修飾グループを置いていく（--group 指定があったサイトのみ）
        for cage, group_name in groups.items():
            direction = self.cages.positions[cage] - self.lattice_sites[site_index]
            direction -= np.floor(direction + 0.5)
            group = place_group(
                direction @ self.cell,
                0.13,  # あとでなんとかする。N-H結合距離
                group_name,
            )
            sites = np.concatenate([sites, group.sites + ion_center])
            atom_names += group.labels
            self.logger.info(f"{atom_names=}")
        return Molecule(
            name=name,
            sites=sites,
            labels=atom_names,
            is_water=False,
        )

    def _effective_spot_cation_groups(self) -> Dict[int, Dict[int, str]]:
        """spot_cation_groups に単位胞由来の cation_groups を展開してマージした辞書を返す（on-demand）。"""
        # ベースは CLI/API で指定された spot_cation_groups
        effective: Dict[int, Dict[int, str]] = {
            k: dict(v) for k, v in self.spot_cation_groups.items()
        }
        # number of nodes in the unitcell
        nuc_nodes = len(self.unitcell.lattice_sites)
        # number of cages in the unitcell
        nuc_cages = len(self.unitcell.cages.specs)
        for un_node, uc_group in self.unitcell.cation_groups.items():
            for node in _replicate_lattice_node(
                un_node, nuc_nodes, self.replication_matrix
            ):
                # 拡大胞のノードに隣接する4つのケージのインデックス
                cage_indices = self.cages.node_to_cage_indices.get(node, [])
                for cage_index in cage_indices:
                    # ケージのインデックスを単位胞でのインデックスになおし、そこに入るグループを取得
                    self.logger.info(f"{cage_index=} {nuc_cages=} {uc_group=} {node=}")
                    uc_cage_index = cage_index % nuc_cages
                    if uc_cage_index in uc_group:
                        group = uc_group[cage_index % nuc_cages]
                        # 拡大胞のノードに隣接する4つのケージのインデックスに、グループをマッピング
                        effective.setdefault(node, {}).setdefault(cage_index, group)
        return effective

    def build_hydronium(self, site_index: int) -> Molecule:
        """H3O+ を構築する。"""
        sites = []
        labels = ["Cn", "Hn", "Hn", "Hn"]
        OH = 0.1
        sites.append(np.zeros(3))
        neighbors = list(self.digraph.successors(site_index))
        if len(neighbors) != 3:
            raise ValueError("H3O+ must have 3 neighbors.")
        for neighbor in neighbors:
            direction = self.lattice_sites[neighbor] - self.lattice_sites[site_index]
            direction -= np.floor(direction + 0.5)
            direction = direction @ self.cell
            direction *= OH / np.linalg.norm(direction)
            sites.append(direction)
        return Molecule(
            name="H3O",
            sites=np.array(sites) + self.lattice_sites[site_index] @ self.cell,
            labels=labels,
            is_water=False,
        )

    def build_hydroxide(self, site_index: int) -> Molecule:
        """OH- を構築する。"""
        sites = []
        labels = ["Nx", "Hx"]
        OH = 0.1
        sites.append(np.zeros(3))
        neighbors = list(self.digraph.successors(site_index))
        if len(neighbors) != 1:
            raise ValueError(f"OH- must have 1 neighbor. {site_index=} {neighbors=}")
        direction = self.lattice_sites[neighbors[0]] - self.lattice_sites[site_index]
        direction -= np.floor(direction + 0.5)
        direction = direction @ self.cell
        direction *= OH / np.linalg.norm(direction)
        sites.append(direction)
        return Molecule(
            name="OH",
            sites=np.array(sites) + self.lattice_sites[site_index] @ self.cell,
            labels=labels,
            is_water=False,
        )

    def substitutional_ions(self) -> Dict[int, Molecule]:
        """単位胞・spot 由来のイオンを統合し、サイト番号→分子の辞書を返す。"""
        # 単位胞由来の group 指定も含めた「実効的な」spot_cation_groups をここで on-demand に組み立てて使う

        # 前提条件: anions, cations, hydroniums, hydroxidesのサイトに重複がないこと。
        # そうでない場合は、エラーを返す。
        if len(set(self.spot_hydroniums) & set(self.spot_hydroxides)) > 0:
            raise ValueError("hydroniums and hydroxides must not have the same site.")
        if len(set(self.anions) & set(self.cations)) > 0:
            raise ValueError("anions and cations must not have the same site.")
        if len(set(self.spot_hydroniums) & set(self.cations)) > 0:
            raise ValueError("hydroniums and cations must not have the same site.")
        if len(set(self.spot_hydroxides) & set(self.cations)) > 0:
            raise ValueError("hydroxides and cations must not have the same site.")

        ions: Dict[int, Molecule] = {}
        for site_index in self.spot_hydroniums:
            ions[site_index] = self.build_hydronium(site_index)
        for site_index in self.spot_hydroxides:
            ions[site_index] = self.build_hydroxide(site_index)
        # ならべかえはここではしない。formatterにまかせる。
        for site_index, molecule in self.anions.items():
            ions[site_index] = self.build_molecular_ion(site_index, molecule)
        effective_groups = self._effective_spot_cation_groups()
        for site_index, molecule in self.cations.items():
            # cationには腕がつく可能性がある。
            groups = effective_groups.get(site_index, {})
            ions[site_index] = self.build_molecular_ion(site_index, molecule, groups)
        self.logger.info(f"{ions=}")
        return ions

    def dope_anions(self, anions: Dict[int, Molecule]) -> None:
        """サイト番号→分子の辞書でアニオン配置を上書きする（主に API/テスト用）。"""
        for site_index, molecule in anions.items():
            self.anions[site_index] = molecule

    def dope_cations(self, cations: Dict[int, Molecule]) -> None:
        """サイト番号→分子の辞書でカチオン配置を上書きする（主に API/テスト用）。"""
        for site_index, molecule in cations.items():
            self.cations[site_index] = molecule

    # def get_atomic_structure(
    #     self,
    #     water_model: Optional[Molecule] = None,
    #     guests: Optional[Dict[str, List[GuestSpec]]] = None,
    #     spot_guests: Optional[Dict[int, Molecule]] = None,
    # ) -> AtomicStructure:
    #     """
    #     原子構造データを統合的に取得する。
    #     exporterプラグインが使用するための統一インターフェース。

    #     Args:
    #         water_model: 水分子モデル（デフォルト: FourSiteWater()）
    #         guests: ケージタイプごとのゲスト分子の指定（デフォルト: {}）
    #         spot_guests: 特定ケージ位置へのゲスト分子の指定（デフォルト: {}）

    #     Returns:
    #         AtomicStructure: 水分子、ゲスト分子、イオン、セル行列を含む統合データ構造
    #     """

    #     return AtomicStructure(
    #         waters=self.water_molecules(water_model=water_model),
    #         guests=self.guest_molecules(
    #             guests=guests or {}, spot_guests=spot_guests or {}
    #         ),
    #         ions=self.substitutional_ions(),
    #         cell=self.cell,
    #     )

    @classmethod
    def get_public_api_properties(cls):
        """
        ユーザー向けAPIとして公開されているpropertyのリストを返す。

        Returns:
            list: 公開API property名のリスト
        """
        return cls.PUBLIC_API_PROPERTIES.copy()

    def list_all_reactive_properties(self):
        """
        すべてのリアクティブプロパティ（DependencyEngineに登録されたタスク）を列挙する。

        Returns:
            dict: property名をキー、タスク関数を値とする辞書
        """
        return self.engine.registry.copy()

    def list_public_reactive_properties(self):
        """
        ユーザー向けAPIとして公開されているリアクティブプロパティのみを列挙する。

        Returns:
            dict: property名をキー、タスク関数を値とする辞書
        """
        all_reactive = self.list_all_reactive_properties()
        public_names = set(self.PUBLIC_API_PROPERTIES)
        return {
            name: func for name, func in all_reactive.items() if name in public_names
        }

    @classmethod
    def list_settable_reactive_properties(cls):
        """setterを持つリアクティブな変数を列挙する"""
        return {
            name: prop
            for name, prop in inspect.getmembers(
                cls, lambda x: isinstance(x, property) and x.fset is not None
            )
        }

    @classmethod
    def list_public_settable_reactive_properties(cls):
        """公開APIのうち、setter を持つリアクティブプロパティのみを列挙する。"""
        return {
            name: prop
            for name, prop in cls.list_settable_reactive_properties().items()
            if name in cls.PUBLIC_API_PROPERTIES
        }



import numpy as np
import genice3.molecule

desc = {"usage": "No options available.", "brief": "An all-atom methane model."}


class Molecule(genice3.molecule.Molecule):
    """
    メタン分子を定義するクラス。
    """

    def __init__(self):
        sites_ = np.array(
            [
                [0.0, 0.0, 0.0],
                [-1.0, -1.0, -1.0],
                [-1.0, +1.0, +1.0],
                [+1.0, -1.0, +1.0],
                [+1.0, +1.0, -1.0],
            ]
        )  # CHHHH
        CH = 0.109  # nm
        sites = sites_ * CH / (3.0**0.5)

        labels = ["C", "H", "H", "H", "H"]
        name = "CH4"
        is_water = False
        super().__init__(sites=sites, labels=labels, name=name, is_water=is_water)




"""
genice3コア機能のテスト
"""

import sys
from pathlib import Path

# プロジェクトのルートディレクトリをPythonパスに追加
project_root = Path(__file__).resolve().parent.parent
if str(project_root) not in sys.path:
    sys.path.insert(0, str(project_root))

import pytest
import numpy as np
from genice3.genice import GenIce3, ConfigurationError
from genice3.plugin import UnitCell


def test_genice3_initialization():
    """GenIce3の初期化テスト"""
    genice = GenIce3()
    assert genice is not None
    assert genice.depol_loop == 1000  # デフォルト値
    # seedは__init__で使用されるが、インスタンス変数として保存されない
    # デフォルト値は1
    genice_with_seed = GenIce3(seed=42)
    assert genice_with_seed is not None


def test_genice3_with_options():
    """オプション付きの初期化テスト"""
    genice = GenIce3(
        depol_loop=500,
        seed=42,
        replication_matrix=np.array([[2, 0, 0], [0, 2, 0], [0, 0, 2]]),
    )
    assert genice.depol_loop == 500
    assert np.array_equal(
        genice.replication_matrix, np.array([[2, 0, 0], [0, 2, 0], [0, 0, 2]])
    )


def test_genice3_unitcell_setting():
    """単位胞の設定テスト"""
    genice = GenIce3()
    genice.unitcell = UnitCell("1h")
    assert genice.unitcell is not None


def test_genice3_invalid_kwargs():
    """無効なキーワード引数のテスト"""
    genice = GenIce3()
    with pytest.raises(ConfigurationError, match="Invalid keyword arguments"):
        genice = GenIce3(invalid_option="test")


def test_genice3_reactive_properties():
    """Reactive propertiesのテスト"""
    genice = GenIce3()
    genice.unitcell = UnitCell("1h")

    # プロパティにアクセスすると計算が実行される
    graph = genice.graph
    assert graph is not None
    assert graph.number_of_nodes() > 0

    digraph = genice.digraph
    assert digraph is not None
    assert digraph.number_of_nodes() == graph.number_of_nodes()


def test_genice3_lattice_sites():
    """格子サイトの取得テスト"""
    genice = GenIce3()
    genice.unitcell = UnitCell("1h")

    lattice_sites = genice.lattice_sites
    assert lattice_sites is not None
    assert len(lattice_sites) > 0
    assert lattice_sites.shape[1] == 3  # 3次元座標


def test_genice3_cell_matrix():
    """セル行列の取得テスト"""
    genice = GenIce3()
    genice.unitcell = UnitCell("1h")

    cell = genice.cell
    assert cell is not None
    assert cell.shape == (3, 3)  # 3x3行列


def test_genice3_water_molecules():
    """水分子の生成テスト"""
    from genice3.plugin import Molecule

    genice = GenIce3()
    genice.unitcell = UnitCell("1h")

    water_model = Molecule("4site")
    waters = genice.water_molecules(water_model=water_model)

    assert len(waters) > 0
    for mol in waters.values():
        assert mol.is_water
        assert len(mol.sites) == len(water_model.sites)


def test_genice3_replication():
    """複製のテスト"""
    replication_matrix = np.array([[2, 0, 0], [0, 2, 0], [0, 0, 2]])
    genice = GenIce3(replication_matrix=replication_matrix)
    genice.unitcell = UnitCell("1h")

    # 複製により原子数が増えることを確認
    from genice3.plugin import Molecule

    water_model = Molecule("4site")
    waters = genice.water_molecules(water_model=water_model)

    # 1hは通常8分子なので、2x2x2で64分子になるはず
    # （実際の値は単位胞の定義による）
    assert len(waters) >= 8  # 最低限元の分子数以上



import genice3.molecule
import numpy as np
import math

desc = {
    "ref": {"TIP3P": "Jorgensen 1983"},
    "brief": "Physical model of water; Oxygen atom is on the lattice point.",
    "usage": "No options available.",
}


class Molecule(genice3.molecule.Molecule):
    """
    物理的な水分子モデルを定義するクラス。酸素原子が格子点上にある。
    """

    def __init__(self):
        oh = 0.09572
        hangle = 104.52 * math.pi / 180 / 2
        mass = 18
        ohz = oh * math.cos(hangle)
        ohy = oh * math.sin(hangle)
        oz = 0  # -ohz*2/mass
        sites = np.array(
            [[0, 0, oz], [0, ohy, ohz + oz], [0, -ohy, ohz + oz]]
        )  # nm, OHHM

        labels = ["O", "H", "H"]
        name = "SOL"
        is_water = True
        super().__init__(sites=sites, labels=labels, name=name, is_water=is_water)



import genice3.group
import numpy as np


class Group(genice3.group.Group):
    def __init__(self):
        bondlen = 0.109
        sine = 1 / 3
        cosine = (1 - sine**2) ** 0.5
        z = bondlen * sine
        r = bondlen * cosine
        cosine2 = -1 / 2
        sine2 = (1 - cosine2**2) ** 0.5
        super().__init__(
            sites=np.array(
                [
                    [0.0, 0.0, 0.0],
                    [r, 0.0, z],
                    [r * cosine2, r * sine2, z],
                    [r * cosine2, -r * sine2, z],
                ]
            ),
            labels=["C", "H", "H", "H"],
            bonds=[(0, 1), (0, 2), (0, 3)],
            name="Methyl",
        )



import sys
from logging import getLogger
from io import TextIOWrapper
from typing import Dict, List, Any, Tuple
from collections import defaultdict

import numpy as np

from cif2ice import cellshape, cellvectors
from genice3.molecule import Molecule
from genice3.genice import GenIce3
from genice3.exporter import parse_water_model_option
format_desc = {
    "suboptions": "water_model: 3site, 4site, 6site, tip4p, etc. (water model name).",
    "aliases": ["lammps", "lmp"],
    "application": "[LAMMPS](https://www.lammps.org/)",
    "extension": ".lammps",
    "water": "Atomic positions",
    "solute": "Atomic positions",
    "hb": "none",
    "remarks": "Yet to be verified.",
}


def _to_lammps_data(
    cellmat: np.ndarray,
    waters: Dict[int, Molecule],
    guests: List[Molecule],
    ions: Dict[int, Molecule],
    command_line: str = None,
) -> str:
    """Output in LAMMPS data format.

    parametersには、hooksで指定したstageの結果が含まれる。
    """
    logger = getLogger("_to_lammps_data")
    logger.info("Generating LAMMPS data file... (Yet to be verified)")

    # セル行列をLAMMPS形式に変換
    # LAMMPSは直交座標系を前提とするため、必要に応じて変換
    if cellmat[0, 1] != 0 or cellmat[0, 2] != 0 or cellmat[1, 2] != 0:
        logger.info(
            "  The specified reshaping matrix does not obey the requirements for LAMMPS' unit cell convention."
        )
        a, b, c, A, B, C = cellshape(cellmat)
        rotmat = np.linalg.inv(cellmat) @ cellvectors(a, b, c, A, B, C)
        logger.info("  The reshape matrix is reoriented.")
    else:
        rotmat = np.eye(3)

    # 分子種ごとにソートする必要がある。
    molecules = defaultdict(list)
    for water in waters.values():
        molecules[water.name].append(water)
    for guest in guests:
        molecules[guest.name].append(guest)
    for ion in ions.values():
        molecules[ion.name].append(ion)

    # 原子タイプのマッピングを作成
    atom_types = {}
    type_counter = 1
    for mol_name in sorted(molecules.keys()):
        mols = molecules[mol_name]
        if len(mols) > 0:
            # 各分子の原子ラベルを取得してタイプを割り当て
            sample_mol = mols[0]
            for label in sample_mol.labels:
                type_key = f"{mol_name}_{label}"
                if type_key not in atom_types:
                    atom_types[type_key] = type_counter
                    type_counter += 1

    # 原子リストを作成
    atoms = []
    molecule_id = 1
    for mol_name in sorted(molecules.keys()):
        mols = molecules[mol_name]
        for mol in mols:
            for name, position in zip(mol.labels, mol.sites):
                type_key = f"{mol_name}_{name}"
                atom_type = atom_types[type_key]
                atoms.append([molecule_id, atom_type, name, position])
            molecule_id += 1

    # セル行列を回転
    cellmat = cellmat @ rotmat

    # LAMMPSデータファイル形式で出力
    s = ""
    if command_line:
        s += f"# {command_line}\n"
    s += "# Generated by GenIce https://github.com/vitroid/GenIce\n"
    s += f"\n{len(atoms)} atoms\n"
    s += f"{len(atom_types)} atom types\n"

    # ボックスサイズを計算
    # LAMMPSは直交座標系を前提とする
    # セルの原点を(0,0,0)として、各軸の範囲を計算
    xlo, xhi = 0.0, cellmat[0, 0]
    ylo, yhi = 0.0, cellmat[1, 1]
    zlo, zhi = 0.0, cellmat[2, 2]
    s += f"\n0.0 {cellmat[0, 0]:.8f} xlo xhi\n"
    s += f"0.0 {cellmat[1, 1]:.8f} ylo yhi\n"
    s += f"0.0 {cellmat[2, 2]:.8f} zlo zhi\n"
    if not (
        abs(cellmat[1, 0]) < 1e-8
        and abs(cellmat[2, 0]) < 1e-8
        and abs(cellmat[2, 1]) < 1e-8
    ):
        # 非直交セル（tilt形式）
        xy = cellmat[1, 0]
        xz = cellmat[2, 0]
        yz = cellmat[2, 1]
        s += f"{xy:.8f} {xz:.8f} {yz:.8f} xy xz yz\n"

    s += "\nAtoms\n\n"

    # 原子座標をnmからAngstromに変換（LAMMPSは通常Angstromを使用）
    # ただし、unitsコマンドで指定できるので、nmのままでも可
    # ここではnmのまま出力（units realまたはunits metalの場合）
    # 電荷は0.0をデフォルトとする（必要に応じて拡張可能）
    charge = 0.0
    for i, atom in enumerate(atoms):
        mol_id, atom_type, atomname, position = atom
        position = position @ rotmat
        s += f"{i + 1:8d} {mol_id:8d} {atom_type:4d} {charge:10.6f} {position[0]:15.8f} {position[1]:15.8f} {position[2]:15.8f}\n"

    logger.info(f"  Total number of atoms: {len(atoms)}")
    logger.info(f"  Total number of atom types: {len(atom_types)}")
    logger.info(f"  Total number of molecules: {molecule_id - 1}")

    return s


def _scalar(v: Any) -> Any:
    if isinstance(v, (list, tuple)) and len(v) == 1:
        return v[0]
    return v


def parse_options(options: Dict[str, Any]) -> Tuple[Dict[str, Any], Dict[str, Any]]:
    """
    lammps プラグインのオプションを処理する。water は water_model のエイリアス。
    """
    options = dict(options)
    if "water" in options and "water_model" not in options:
        options["water_model"] = options["water"]
    processed: Dict[str, Any] = {}
    unprocessed = dict(options)
    if "water_model" in options:
        processed["water_model"] = _scalar(options["water_model"])
        unprocessed.pop("water_model", None)
    return processed, unprocessed


def dumps(
    genice: GenIce3,
    water_model: str = "3site",
    command_line: str = "",
    name: str = "",  # dummy
    **kwargs,
):
    """
    LAMMPS形式で出力

    Args:
        genice: GenIce3インスタンス
        water_model: 水分子モデル名
        command_line: コマンドライン文字列
        name: プラグイン名（"lammps"または""）
    """
    logger = getLogger("lammps.dump")
    if name not in ("lammps", ""):
        raise ValueError(f"name must be 'lammps' or '', got: {name}")
    water_model = parse_water_model_option(water_model)
    waters = genice.water_molecules(water_model=water_model)
    guests = genice.guest_molecules()
    ions = genice.substitutional_ions()

    return _to_lammps_data(
        cellmat=genice.cell,
        waters=waters,
        guests=guests,
        ions=ions,
        command_line=command_line,
    )


def dump(genice: GenIce3, file: TextIOWrapper = sys.stdout, **options):
    file.write(dumps(genice, **options))



# coding: utf-8
"""
Crude Cif file format
"""

from logging import getLogger
from io import TextIOWrapper
import sys
from typing import Dict, List, Any, Tuple

import numpy as np

from cif2ice import cellshape
from genice3.genice import GenIce3
from genice3.exporter import parse_water_model_option

format_desc = {
    "aliases": ["cif"],
    "application": "CIF",
    "extension": ".cif",
    "water": "Atomic positions",
    "solute": "Atomic positions",
    "hb": "none",
    "remarks": "Experimental",
}


# def parse_options(options: Dict[str, Any]) -> Tuple[Dict[str, Any], Dict[str, Any]]:
# オプションがないので関数自体を準備しない。呼ぶ側が判断してくれ。


def _format_cell_shape(a, b, c, A, B, C):
    aL, bL, cL, alpha, beta, gamma = a, b, c, A, B, C

    s = ""
    s += "data_genice\n"
    s += "#" + "\n#Command line: " + " ".join(sys.argv) + "\n"
    s += "_cell_length_a                {0}\n".format(aL * 10)
    s += "_cell_length_b                {0}\n".format(bL * 10)
    s += "_cell_length_c                {0}\n".format(cL * 10)
    s += "_cell_angle_alpha             {0}\n".format(alpha)
    s += "_cell_angle_beta              {0}\n".format(beta)
    s += "_cell_angle_gamma             {0}\n".format(gamma)
    s += "\n"
    rights = np.array([alpha, beta, gamma]) - 90
    if np.allclose(rights, 0):
        s += "_symmetry_cell_setting        'orthorhombic'\n"
        s += "_symmetry_space_group_name_H-M   'P 1 '\n"
    else:
        # for now it is always triclinic
        s += "_symmetry_cell_setting        'triclinic'\n"
        s += "_symmetry_space_group_name_H-M   'P 1 '\n"
    return s


def dump(
    genice: GenIce3,
    file: TextIOWrapper = sys.stdout,
    water_model: str = "3site",
    name: str = "",
    command_line: str = "",
    **kwargs,
):
    "Output in CIF format."
    logger = getLogger()
    if name not in ("cif", ""):
        raise ValueError(f"name must be 'cif' or '', got: {name}")

    a, b, c, A, B, C = cellshape(genice.cell)
    s = _format_cell_shape(a, b, c, A, B, C)

    s += """
loop_
_symmetry_equiv_pos_as_xyz
X,Y,Z

loop_
_atom_site_label
_atom_site_type_symbol
_atom_site_fract_x
_atom_site_fract_y
_atom_site_fract_z
"""
    # オプションの処理
    water_model = parse_water_model_option(water_model)

    # 分子の取得
    waters = genice.water_molecules(water_model=water_model)
    guests = genice.guest_molecules()
    ions = genice.substitutional_ions()

    # セル行列の逆行列を計算（絶対座標から相対座標への変換用）
    cell_matrix = genice.cell
    cell_inv = np.linalg.inv(cell_matrix)

    # 原子を収集
    atoms = []
    for water in waters.values():
        for name, position in zip(water.labels, water.sites):
            atoms.append([water.name, name, position])
    for guest in guests:
        for name, position in zip(guest.labels, guest.sites):
            atoms.append([guest.name, name, position])
    for ion in ions.values():
        for name, position in zip(ion.labels, ion.sites):
            atoms.append([ion.name, name, position])

    # CIF形式で出力
    for i, atom in enumerate(atoms):
        resname, atomname, position = atom
        # 絶対座標から相対座標に変換
        pos = position @ cell_inv
        pos = pos - np.floor(pos)  # 0-1の範囲に正規化
        label = f"{atomname}{i}"
        s += f"{label:>6}{atomname:>6}{pos[0]:10.4f}{pos[1]:10.4f}{pos[2]:10.4f}\n"
    s += "\n"
    file.write(s)



"""
option_parser を使った CLI 実行フロー。
argv → パース → プラグイン dispatch → result（genice.py が期待する形）。
"""

from __future__ import annotations

from pathlib import Path
from typing import Any, Dict, List, Tuple

from genice3.cli.option_parser import (
    parse_options,
    scalarize_single_item_lists,
    structure_for_display,
)
from genice3.cli.options import (
    base_options_from_new_structure,
    get_common_unitcell_option_names,
    get_short_to_long_option_names,
)
from genice3.plugin import safe_import

try:
    import yaml
except ImportError:
    yaml = None

# 基底オプションのキー（runner で unitcell/exporter に渡さない）
_BASE_KEYS = frozenset(
    {
        "debug",
        "seed",
        "rep",
        "replication_factors",
        "replication_matrix",
        "depol_loop",
        "target_polarization",
        "config",
        "exporter",
        "guest",
        "spot_guest",
        "spot_anion",
        "spot_cation",
    }
)


def load_config_file(yaml_path: str) -> Dict[str, Any]:
    """YAML 設定を読み、option_parser と同じ構造で返す。"""
    if yaml is None:
        return {}
    p = Path(yaml_path)
    if not p.exists():
        return {}
    with open(p, "r", encoding="utf-8") as f:
        config = yaml.safe_load(f) or {}
    # トップに unitcell / genice3 等がある形式を、フラットな option 構造に正規化
    out: Dict[str, Any] = {}
    if "unitcell" in config:
        uc = config["unitcell"]
        if isinstance(uc, dict):
            out["unitcell"] = uc.get("name", "")
            for k, v in uc.items():
                if k != "name":
                    out[k] = v
        else:
            out["unitcell"] = uc
    if "genice3" in config:
        out.update(config["genice3"])
    if "exporter" in config:
        out["exporter"] = config["exporter"]
    # その他のトップレベルキー（unitcell 用 file, osite や shift, density 等）も渡す
    for k, v in config.items():
        if k not in out:
            out[k] = v
    return out


def _merge_config_cmdline(config: Dict[str, Any], cmdline: Dict[str, Any]) -> Dict[str, Any]:
    """設定ファイルをコマンドラインで上書き。"""
    out = dict(config)
    for k, v in cmdline.items():
        if v is not None and (k != "unitcell" or v != ""):
            out[k] = v
    return out


def _get_exporter_name_and_options(data: Dict[str, Any]) -> Tuple[str, Dict[str, Any]]:
    """data["exporter"] から (name, subopts_dict) を返す。"""
    raw = data.get("exporter")
    if raw is None:
        return "gromacs", {}
    items = raw if isinstance(raw, list) else [raw]
    if not items:
        return "gromacs", {}
    first = items[0]
    if isinstance(first, dict):
        (name, subopts), = first.items()
        return str(name), dict(subopts)
    return str(first), {}


def parse_argv(argv: List[str]) -> Dict[str, Any]:
    """
    argv（sys.argv[1:]）をパースし、genice.py の get_result() と同じ形の辞書を返す。
    """
    args = list(argv)
    config: Dict[str, Any] = {}
    i = 0
    while i < len(args):
        if args[i] in ("--config", "-Y"):
            i += 1
            if i < len(args):
                config = load_config_file(args[i])
            i += 1
        else:
            i += 1

    # --config 以外を並べて option_parser に渡す
    line_parts = []
    i = 0
    while i < len(argv):
        if argv[i] in ("--config", "-Y"):
            i += 2
            continue
        line_parts.append(argv[i])
        i += 1
    line = " ".join(line_parts)

    if not line_parts or line_parts[0].startswith("-") or line_parts[0].startswith(":"):
        # コマンドラインに unitcell が無くても --config で読んだ設定があれば使う
        if config and config.get("unitcell"):
            merged = _merge_config_cmdline(config, {})
        else:
            return {
                "base_options": {},
                "unitcell": {"name": None, "options": {}, "processed": {}},
                "exporter": {"name": "gromacs", "options": {}, "processed": {}},
                "plugin_chain": [],
            }
    else:
        try:
            parsed = parse_options(line)
        except ValueError as e:
            raise RuntimeError(f"オプションのパースに失敗しました: {e}") from e
        # 短いオプション名 (-e → exporter 等) を long 名に正規化
        for short, long_name in get_short_to_long_option_names().items():
            if short in parsed:
                parsed.setdefault(long_name, parsed.pop(short))
        display = structure_for_display(parsed)
        merged = _merge_config_cmdline(config, display)

    unitcell_name = str(merged.get("unitcell", ""))
    if not unitcell_name:
        return {
            "base_options": {},
            "unitcell": {"name": None, "options": {}, "processed": {}},
            "exporter": {"name": "gromacs", "options": {}, "processed": {}},
            "plugin_chain": [],
        }

    # 基底オプション
    base_options = base_options_from_new_structure(merged)

    # unitcell に渡すオプション（新構造のまま）
    unitcell_options = {
        k: merged[k]
        for k in merged
        if k not in _BASE_KEYS and k not in ("unitcell", "exporter")
    }

    # unitcell プラグインの parse_options（新構造を受け取る）
    # モジュール級 parse_options を優先（CIF の file/osite 等はこちらで処理）
    unitcell_processed: Dict[str, Any] = {}
    unitcell_unprocessed: Dict[str, Any] = {}
    try:
        uc_module = safe_import("unitcell", unitcell_name)
        if hasattr(uc_module, "parse_options"):
            unitcell_processed, unitcell_unprocessed = uc_module.parse_options(
                unitcell_options
            )
        elif hasattr(uc_module, "UnitCell") and hasattr(
            uc_module.UnitCell, "parse_options"
        ):
            unitcell_processed, unitcell_unprocessed = uc_module.UnitCell.parse_options(
                unitcell_options
            )
        else:
            # parse_options がない unitcell: 共通オプションのみ消費、残りは unprocessed（警告対象）
            common = get_common_unitcell_option_names()
            unitcell_processed = {
                k: v for k, v in unitcell_options.items() if k in common
            }
            unitcell_unprocessed = {
                k: v for k, v in unitcell_options.items() if k not in common
            }
    except Exception:
        # 読み込み失敗時も同様に共通のみ消費、残りは unprocessed
        common = get_common_unitcell_option_names()
        unitcell_processed = {
            k: v for k, v in unitcell_options.items() if k in common
        }
        unitcell_unprocessed = {
            k: v for k, v in unitcell_options.items() if k not in common
        }

    # exporter
    exporter_name, exporter_subopts = _get_exporter_name_and_options(merged)
    exporter_processed: Dict[str, Any] = {}
    exporter_unprocessed: Dict[str, Any] = {}
    try:
        ex_module = safe_import("exporter", exporter_name)
        if hasattr(ex_module, "parse_options"):
            exporter_processed, exporter_unprocessed = ex_module.parse_options(
                exporter_subopts
            )
    except Exception:
        pass

    return {
        "base_options": base_options,
        "unitcell": {
            "name": unitcell_name,
            "options": unitcell_options,
            "processed": unitcell_processed,
            "unprocessed": unitcell_unprocessed,
        },
        "exporter": {
            "name": exporter_name,
            "options": exporter_subopts,
            "processed": exporter_processed,
            "unprocessed": exporter_unprocessed,
        },
        "plugin_chain": [],
    }


def validate_result(result: Dict[str, Any]) -> Tuple[bool, List[str]]:
    """unitcell 名が指定されているか検証。"""
    errors = []
    if not result.get("unitcell", {}).get("name"):
        errors.append("unitcell名が指定されていません")
    return (len(errors) == 0, errors)



# coding: utf-8

import itertools as it
import sys
from collections import defaultdict
from logging import getLogger
from io import TextIOWrapper

import numpy as np
import yaplotlib as yp

from genice3.genice import GenIce3
from genice3.exporter import parse_water_model_option
from genice3.util import serialize

desc = {
    "ref": {"Codes": "https://github.com/vitroid/Yaplot"},
    "brief": "Yaplot.",
    "usage": """
Usage: genice3 UNITCELL --exporter yaplot [:option value ...]

options:
    H=x   Set the radius of H to be x.
""",
}

format_desc = {
    "aliases": ["y", "yaplot"],
    "application": "[Yaplot](https://github.com/vitroid/Yaplot)",
    "extension": ".yap",
    "water": "Atomic positions",
    "solute": "Atomic positions",
    "hb": "o",
    "remarks": "It renders molecular configurations and the HB network.",
    "suboptions": "H: radius of H atom (numeric).",
}


def dump(genice: GenIce3, file: TextIOWrapper = sys.stdout, **options):
    "Draw the cell in Yaplot format."
    logger = getLogger()

    size_O = 0.03
    size_H = float(options.get("H", 0)) * size_O

    s = yp.Layer(2)
    x, y, z = genice.cell
    for p, q, r in ((x, y, z), (y, z, x), (z, x, y)):
        for a in (np.zeros(3), p, q, p + q):
            s += yp.Line(a, a + r)

    if size_H == 0:
        # 水素のサイズ指定がない場合
        # prepare the reverse dict
        waters = defaultdict(dict)
        pos = genice.lattice_sites
        s += yp.Layer(4)
        s += yp.Color(3)
        s += yp.Size(0.03)
        for p in pos:
            s += yp.Circle(p @ genice.cell)
        s += yp.Layer(5)
        s += yp.Color(4)
        # s += yp.Size(0.03)
        for i, j in genice.graph.edges(data=False):
            # if i in waters and j in waters:  # edge may connect to the dopant
            O1, O2 = pos[i], pos[j]
            d = O2 - O1
            d -= np.floor(d + 0.5)
            O2 = O1 + d
            s += yp.Line(O1 @ genice.cell, O2 @ genice.cell)

    # waterとwater_modelの両方をサポート（後方互換性のため）
    water_model_name = options.get("water_model") or options.get("water", "4site")
    water_model = parse_water_model_option(water_model_name)

    waters = genice.water_molecules(water_model=water_model)
    guests = genice.guest_molecules()
    ions = genice.substitutional_ions()

    atoms = serialize(list(waters.values()))

    logger.info("  Total number of atoms: {0}".format(len(atoms)))

    # prepare the reverse dict
    water_sites = defaultdict(dict)
    for water_index, water in waters.items():
        for atom_name, position in zip(water.labels, water.sites):
            if "O" in atom_name:
                water_sites[water_index]["O"] = position
            elif "H" in atom_name:
                if "H0" not in water_sites[water_index]:
                    water_sites[water_index]["H0"] = position
                else:
                    water_sites[water_index]["H1"] = position
    s += yp.Color(3)
    for water_index, water in water_sites.items():
        O = water["O"]
        H0 = water["H0"]
        H1 = water["H1"]
        s += yp.Layer(4)
        s += yp.Color(3)
        s += yp.Size(size_O)
        s += yp.Circle(O)

        s += yp.Line(O, H0)
        s += yp.Line(O, H1)

        s += yp.Size(size_H)
        s += yp.Circle(H0)
        s += yp.Circle(H1)

        s += yp.Color(2)
        s += yp.Layer(1)
        s += yp.Text(O, f"{water_index}")

    s += yp.Layer(3)
    s += yp.Color(4)
    s += yp.ArrowType(1)
    s += yp.Size(size_O)
    for i, j in genice.digraph.edges(data=False):
        if i in waters and j in waters:  # edge may connect to the dopant
            O = water_sites[j]["O"]
            H0 = water_sites[i]["H0"]
            H1 = water_sites[i]["H1"]
            d0 = H0 - O
            d1 = H1 - O
            rr0 = d0 @ d0
            rr1 = d1 @ d1
            if rr0 < rr1 and rr0 < 0.245**2:
                s += yp.Arrow(H0, O)
            if rr1 < rr0 and rr1 < 0.245**2:
                s += yp.Arrow(H1, O)

    gatoms = serialize(guests) + serialize(list(ions.values()))
    palettes = dict()

    s += yp.Layer(4)
    s += yp.ArrowType(1)
    H = []
    O = ""
    for molecule_name, atom_name, position in gatoms:
        if atom_name in palettes:
            pal = palettes[atom_name]
        else:
            pal = 4 + len(palettes)
            palettes[atom_name] = pal
        s += yp.Color(pal)
        if atom_name[0] == "H":
            s += yp.Size(size_H)
        else:
            s += yp.Size(size_O)
        s += yp.Circle(position)
    for a, b in it.combinations(gatoms, 2):
        resname, atomname, position1 = a
        resname, atomname, position2 = b
        d = position1 - position2
        if d @ d < 0.16**2:
            s += yp.Line(position1, position2)

    cagespecs = genice.cages

    s += yp.Layer(6)
    s += yp.Size(size_O * 3)
    for i, (cagespec, position) in enumerate(zip(cagespecs.specs, cagespecs.positions)):
        s += yp.Color(5)
        s += yp.Circle(position @ genice.cell)
        s += yp.Color(2)
        s += yp.Text(position @ genice.cell, f"{i}")

        # for atom in cagespec.graph:
        #     s += yp.Line(position, atom)

    s += "#" + "\n#Command line: " + " ".join(sys.argv) + "\n"
    s += yp.NewPage()
    file.write(s)



desc = {
    "ref": {},
    "brief": "Half lattice of ice VI.",
    "test": ({"options": "-r 2 2 2"},),
}

import genice3.unitcell
import numpy as np
from cif2ice import cellvectors


class UnitCell(genice3.unitcell.UnitCell):
    """
    6h単位胞を定義するクラス。
    """

    def __init__(self, **kwargs):
        graph = None  # pairsがない場合は自動生成

        waters = np.fromstring(
            """
        0.2200    0.5000    0.3800
        0.7800    0.5000    0.3800
        0.5000    0.2200    0.6200
        0.5000    0.5000    0.0000
        0.5000    0.7800    0.6200
        """,
            sep=" ",
        ).reshape(-1, 3)

        coord = "relative"

        bondlen = 2.3681227356441177

        density = 0.6865

        cell = cellvectors(a=4.87672629, b=4.87385128, c=4.49131038)

        super().__init__(
            cell=cell,
            lattice_sites=waters,
            coord=coord,
            bondlen=bondlen,
            density=density,
            **kwargs,
        )



"""OptionDef と型定数（--help と検証用）。options と validator の両方から参照する。"""

from dataclasses import dataclass
from typing import Any, Callable, Optional

OPTION_TYPE_FLAG = "flag"
OPTION_TYPE_STRING = "string"
OPTION_TYPE_TUPLE = "tuple"
OPTION_TYPE_KEYVALUE = "keyvalue"


@dataclass(frozen=True)
class OptionDef:
    """CLI で認識するオプションの定義（ヘルプ・検証用）。"""

    name: str
    short: Optional[str] = None
    is_flag: bool = False
    max_values: Optional[int] = None
    level: str = "base"
    parse_type: Optional[str] = None
    metavar: Optional[str] = None
    help_text: Optional[str] = None
    help_format: Optional[str] = None
    parse_validator: Optional[Callable[[Any], None]] = None



import numpy as np
import genice3.molecule


class Molecule(genice3.molecule.Molecule):
    """
    単一サイトの分子を定義するクラス。
    """

    def __init__(self, label: str = "X", name: str = "X"):
        sites = np.array([[0, 0, 0]])
        labels = [
            label,
        ]
        is_water = False
        super().__init__(sites=sites, labels=labels, name=name, is_water=is_water)



"""
genice3 CLI のオプション検証

パース直後に実行する parse_validator を定義。
runtime_validator（構造体などコンテキストが必要）は必要になったら追加。
"""

from typing import Any, Dict, Sequence

from genice3.cli.option_def import OptionDef


# =============================================================================
# parse_validator 関数（パース直後に実行、コンテキスト不要）
# =============================================================================


def validate_seed(value: Any) -> None:
    """seed は非負整数であること"""
    if value is None:
        return
    try:
        v = int(value)
        if v < 0:
            raise ValueError(f"--seed must be non-negative, got: {v}")
    except (TypeError, ValueError) as e:
        if isinstance(e, ValueError) and "non-negative" in str(e):
            raise
        raise ValueError(f"--seed must be an integer, got: {value}") from e


def validate_replication_factors(value: Any) -> None:
    """replication_factors は 1〜3 個の正の整数であること"""
    if value is None:
        return
    t = value if isinstance(value, (tuple, list)) else (value,)
    if len(t) < 1 or len(t) > 3:
        raise ValueError(
            f"--rep / --replication_factors must have 1–3 values, got {len(t)}"
        )
    for v in t:
        try:
            n = int(v)
            if n < 1:
                raise ValueError(
                    f"--rep / --replication_factors values must be positive, got: {v}"
                )
        except (TypeError, ValueError) as e:
            if isinstance(e, ValueError) and "positive" in str(e):
                raise
            raise ValueError(
                f"--rep / --replication_factors must be integers, got: {v}"
            ) from e


def validate_depol_loop(value: Any) -> None:
    """depol_loop は正の整数であること"""
    if value is None:
        return
    try:
        n = int(value)
        if n < 1:
            raise ValueError(f"--depol_loop must be positive, got: {n}")
    except (TypeError, ValueError) as e:
        if isinstance(e, ValueError) and "positive" in str(e):
            raise
        raise ValueError(f"--depol_loop must be an integer, got: {value}") from e


def validate_target_polarization(value: Any) -> None:
    """target_polarization は3個の数値であること"""
    if value is None:
        return
    if hasattr(value, "shape") and value.shape == (3,):
        return
    t = value if isinstance(value, (tuple, list)) else (value,)
    if len(t) != 3:
        raise ValueError(
            f"--target_polarization must have exactly 3 values, got: {len(t)}"
        )
    for v in t:
        try:
            float(v)
        except (TypeError, ValueError):
            raise ValueError(
                f"--target_polarization must be numbers, got: {value}"
            ) from None


def validate_replication_matrix(value: Any) -> None:
    """replication_matrix は9個の整数であること"""
    if value is None:
        return
    t = value if isinstance(value, (tuple, list)) else (value,)
    if hasattr(value, "size"):
        t = list(value.flatten())
    if len(t) != 9:
        raise ValueError(
            f"--replication_matrix must have exactly 9 integers, got: {len(t)}"
        )
    for v in t:
        try:
            int(v)
        except (TypeError, ValueError):
            raise ValueError(
                f"--replication_matrix must be integers, got: {value}"
            ) from None


def validate_spot_ion_dict(value: Any, option_name: str) -> None:
    """spot_cation / spot_anion の辞書: キーは非負整数、または '51=N' 形式（= の前がサイト番号）"""
    if value is None or (isinstance(value, dict) and not value):
        return
    if not isinstance(value, dict):
        raise ValueError(
            f"--{option_name} must be key=value format or bracketed, got: {type(value).__name__}"
        )
    for k in value.keys():
        key_str = str(k).strip()
        # '51=N' 形式なら = の前をサイト番号として検証
        if "=" in key_str:
            key_str = key_str.split("=", 1)[0].strip()
        try:
            idx = int(key_str)
            if idx < 0:
                raise ValueError(
                    f"--{option_name} site index must be non-negative, got: {k}"
                )
        except (TypeError, ValueError) as e:
            if isinstance(e, ValueError) and "non-negative" in str(e):
                raise
            raise ValueError(
                f"--{option_name} site index must be an integer (or LABEL=ION), got: {k}"
            ) from e


# =============================================================================
# 一括検証
# =============================================================================


def validate_parsed_options(
    base_options: Dict[str, Any],
    option_defs: Sequence[OptionDef],
) -> None:
    """
    パース済み base_options に対して parse_validator を実行する。
    エラーがあれば ValueError を raise する。
    """
    for def_ in option_defs:
        if def_.parse_validator is None:
            continue
        # rep は build_options_dict で replication_factors に正規化される
        key = "replication_factors" if def_.name == "rep" else def_.name
        if key not in base_options:
            continue
        def_.parse_validator(base_options[key])



"""
Utility functions for genice3
"""

from dataclasses import dataclass
from typing import List, Iterable, Union, Tuple
from logging import getLogger
from collections import defaultdict
import itertools as it

import numpy as np
import networkx as nx
import pairlist as pl

from genice3.molecule import Molecule


def replicate_positions(positions1, replica_vectors, grand_cellmat):
    """レプリカ単位胞の数だけ、水分子位置を複製する。"""
    # 空の配列の場合は空の配列を返す
    if len(positions1) == 0:
        return np.array([]).reshape(0, 3)
    inv = np.linalg.inv(grand_cellmat)
    reppositions = []
    for replica_vector in replica_vectors:
        # レプリカ単位胞内での分子の位置
        replica = positions1 + replica_vector
        # 大セル内の位置に換算する
        replica = replica @ inv
        replica -= np.floor(replica)
        # 束ねて
        reppositions.append(replica)
    # 一つの配列にまとめ
    reppositions = np.vstack(reppositions)
    return reppositions


def grandcell_wrap(
    cell1frac_vec: np.ndarray, reshape: np.ndarray, invdet: np.ndarray, det: int
):
    """
    単位胞の小数座標で表された点(単位胞内とは限らない)を、拡大単位胞内におさめる。
    例えば、拡大単位胞が(-1,4), (6,1)なら、(5.1,5.1)と(6.1,1.1)は(0.1,0.1)になる。
    """
    frac = cell1frac_vec @ invdet
    frac = frac % det
    return frac @ reshape / det


def assume_tetrahedral_vectors(v):
    """
    Assume missing vectors at a tetrahedral node.

    Given: known vectors.
    Returns: assumed vectors
    """
    if len(v) == 0:
        raise ValueError("at least one known vector is required")

    if len(v) == 3:
        return [-(v[0] + v[1] + v[2])]

    if len(v) == 2:
        y = v[1] - v[0]
        y /= np.linalg.norm(y)
        z = v[1] + v[0]
        z /= np.linalg.norm(z)
        x = np.cross(y, z)
        v2 = (x * 8.0**0.5 - z) / 3.0
        v3 = (-x * 8.0**0.5 - z) / 3.0
        return [v2, v3]

    if len(v) == 1:
        vr = np.random.rand(3)
        vr /= np.linalg.norm(vr)
        z = v[0] / np.linalg.norm(v[0])
        x = np.cross(z, vr)
        y = np.cross(z, x)
        x1 = -x / 2 + 3.0**0.5 * y / 2
        x2 = -x / 2 - 3.0**0.5 * y / 2
        return [x, x1, x2]

    return []


def serialize(molecules: List[Molecule]):
    atoms = []
    for molecule in molecules:
        for name, position in zip(molecule.labels, molecule.sites):
            atoms.append([molecule.name, name, position])
    return atoms


# ============================================================================
# CIF-like data processing functions
# ============================================================================


def fullatoms(atomd, sops):
    """Generate all atoms from atom dictionary and symmetry operations."""
    global x, y, z
    logger = getLogger()
    # the variables to be evaluated by eval() must be global (?)
    atoms = []
    for sop in sops:
        for name, pos in atomd.items():
            x, y, z = pos
            p = sop(x, y, z)
            tooclose = False
            for n, f in atoms:
                d = f - p
                d -= np.floor(d + 0.5)
                L2 = d @ d
                if L2 < 0.0001:
                    logger.debug("Too close: {0} {1}".format(f, p))
                    tooclose = True
                    break
            if not tooclose:
                atoms.append((name, p))
    return atoms


def atomdic(atoms):
    """Parse atom positions from a string format."""
    atomd = dict()
    for atom in atoms.split("\n"):
        cols = atom.split()
        if len(cols) == 0:
            continue
        # remove error digits from 1-3 cols.
        xyz = []
        for col in cols[1:4]:
            p = col.find("(")
            if p >= 0:
                col = col[:p]
            xyz.append(float(col))
        name = cols[0]
        atomd[name] = xyz
    return atomd


def symmetry_operators(symops: str, offsets: Iterable = [("+0", "+0", "+0")]):
    """Generator of the symmetry operations

    Args:
        symops (str): The table of symmetry operators, e.g. http://img.chem.ucl.ac.uk/sgp/large/036az3.htm
        offset (Iterable, optional): Offset to the origin. Defaults to ["+0", "+0", "+0"].
    """

    def _wrap(x):
        return x
        return x - np.floor(x + 0.5)

    def eval_(v, x, y, z):
        return eval(v)

    symops = symops.translate(str.maketrans("XYZ", "xyz"))
    for symop in symops.split("\n"):
        cols = symop.split(",")
        if len(cols) <= 1:
            cols = symop.split()
            if len(cols) <= 1:
                continue
        for offset in offsets:
            ops = []
            for col, offs in zip(cols, offset):
                ops.append(col + offs)
            yield lambda x, y, z: _wrap(np.array([eval_(op, x, y, z) for op in ops]))


# def atoms_to_waters(oxygens, hydrogens, cell, partial_order=False):
#     logger = getLogger("atoms_to_waters")
#     oh = defaultdict(list)
#     parent = dict()
#     # find covalent OH bonds
#     for i, j in pl.pairs_iter(
#         oxygens, maxdist=0.15, cell=cell, pos2=hydrogens, distance=False  # nm
#     ):
#         oh[i].append(j)
#         parent[j] = i
#     logger.debug(parent)
#     pairs = []
#     # find HBs
#     for i, j in pl.pairs_iter(
#         oxygens, maxdist=0.20, cell=cell, pos2=hydrogens, distance=False  # nm
#     ):
#         if j not in oh[i]:
#             # H is on a different water molecule
#             p = parent[j]
#             pairs.append([p, i])

#     logger.debug(pairs)

#     if partial_order:
#         oo_pairs = [
#             (i, j)
#             for i, j in pl.pairs_iter(oxygens, maxdist=0.30, cell=cell, distance=False)
#         ]
#         # positions of the oxygen atoms, fixed edges, and unassigned edges.
#         return oxygens, pairs, oo_pairs

#     waters = []
#     for i in range(len(oh)):
#         logger.debug((i, oh[i]))
#         j, k = oh[i]
#         dhj = hydrogens[j] - oxygens[i]
#         dhk = hydrogens[k] - oxygens[i]
#         dhj -= np.floor(dhj + 0.5)
#         dhk -= np.floor(dhk + 0.5)
#         com = oxygens[i] + (dhj + dhk) / 18
#         waters.append(com)

#     return waters, pairs


def atoms_to_waters(oxygens, hydrogens, cell, partial_order=False):
    """
    原子座標 (O, H) から水分子の重心座標と水素結合ペアを求める。

    Parameters
    ----------
    oxygens : np.ndarray
        O原子の分数座標 (shape: (n_O, 3))
    hydrogens : np.ndarray
        H原子の分数座標 (shape: (n_H, 3))
    cell : np.ndarray
        セル行列
    partial_order : bool, default False
        True のときは、重心計算をせず、O–O ペアも返す（秩序化用）。

    Returns
    -------
    if partial_order is False:
        waters : list[np.ndarray]
            水分子の重心座標のリスト
        pairs : list[list[int, int]]
            水素結合ペア (donor_O_index, acceptor_O_index) のリスト

    if partial_order is True:
        oxygens : np.ndarray
        pairs   : list[list[int, int]]  (O–O 間のHB由来ペア)
        oo_pairs: list[tuple[int, int]] (距離しきい値以内の O–O ペア)
    """
    logger = getLogger("atoms_to_waters")

    # --- Step 1: 共有結合 OH（親Oを決める） ------------------------------------
    oh = defaultdict(list)  # Oインデックス -> 共有結合しているHインデックスのリスト
    parent = {}  # Hインデックス -> 親Oインデックス

    for o_idx, h_idx in pl.pairs_iter(
        oxygens, maxdist=0.15, cell=cell, pos2=hydrogens, distance=False  # nm
    ):
        oh[o_idx].append(h_idx)
        parent[h_idx] = o_idx

    logger.debug("covalent OH parent map: %s", parent)

    # --- Step 2: 水素結合ペア (O_donor, O_acceptor) を集める --------------------
    pairs = []
    for o_idx, h_idx in pl.pairs_iter(
        oxygens, maxdist=0.20, cell=cell, pos2=hydrogens, distance=False  # nm
    ):
        # 共有結合しているHは除外（別の水分子に属するHだけを使う）
        if h_idx not in oh[o_idx]:
            donor = parent[h_idx]  # このHの親O
            acceptor = o_idx
            pairs.append([donor, acceptor])

    logger.debug("HB pairs (donor, acceptor): %s", pairs)

    # --- Step 3: 秩序化用の部分情報だけ欲しい場合 ------------------------------
    if partial_order:
        oo_pairs = [
            (i, j)
            for i, j in pl.pairs_iter(oxygens, maxdist=0.30, cell=cell, distance=False)
        ]
        # oxygens: O原子座標
        # pairs  : HBペア (donor, acceptor)
        # oo_pairs: 近接 O–O ペア
        return oxygens, pairs, oo_pairs

    # --- Step 4: 各Oごとに水分子重心を計算 ------------------------------------

    def _water_com(o_index, h_indices):
        """1つのOと2つのHから、水分子重心の分数座標を計算する。"""
        h1_index, h2_index = h_indices

        o = oxygens[o_index]
        h1 = hydrogens[h1_index]
        h2 = hydrogens[h2_index]

        # 近接イメージを選ぶ（周期境界考慮）
        dh1 = h1 - o
        dh2 = h2 - o
        dh1 -= np.floor(dh1 + 0.5)
        dh2 -= np.floor(dh2 + 0.5)

        # O質量:16, H質量:1 として COM = (16*O + H1 + H2)/18
        return o + (dh1 + dh2) / 18.0

    waters = []
    for o_index in range(len(oh)):
        logger.debug("O %d has Hs %s", o_index, oh[o_index])
        # ここでは H がちょうど2個ある前提（既存コードと同じ前提）
        waters.append(_water_com(o_index, oh[o_index]))

    return waters, pairs


def waters_and_pairs(
    cell,
    atomd,
    sops,
    rep=(1, 1, 1),
    O_labels=("O",),
    H_labels="DH",
    partial_order=False,
):
    """Generate water positions and pairs from CIF-like data.

    Args:
        cell: Cell matrix
        atomd: Atom dictionary from atomdic()
        sops: Symmetry operators from symmetry_operators()
        rep: Replication factors (default: (1, 1, 1))
        O_labels: Labels for oxygen atoms (default: ("O",))
        H_labels: Labels for hydrogen atoms (default: "DH")
        partial_order: Whether to return partial order information (default: False)

    Returns:
        waters: Water positions (numpy array)
        pairs: Hydrogen bond pairs (list or None)
        oo_pairs: O-O pairs (only if partial_order=True)
    """
    # 部分秩序の場合は、位置が確定している水素だけをatomdに入れ、未確定の水素は省く。
    logger = getLogger()
    oxygens = []
    hydrogens = []
    for name, pos in fullatoms(atomd, sops):
        if name[0] in O_labels:
            oxygens.append(pos)
        elif name[0] in H_labels:
            hydrogens.append(pos)

    if not partial_order:
        if len(oxygens) * 2 != len(hydrogens) and len(hydrogens) != 0:
            raise ValueError(
                f"hydrogen count must be 2×oxygen count or 0: "
                f"O={len(oxygens)}, H={len(hydrogens)}"
            )
    cell *= np.array(rep)

    oo = [
        [o[0] + x, o[1] + y, o[2] + z]
        for o in oxygens
        for x in range(rep[0])
        for y in range(rep[1])
        for z in range(rep[2])
    ]
    oxygens = np.array(oo)
    oxygens /= np.array(rep)

    # if no hydrogens are included in atomd,
    if len(hydrogens) == 0:
        return oxygens, None

    hh = [
        [h[0] + x, h[1] + y, h[2] + z]
        for h in hydrogens
        for x in range(rep[0])
        for y in range(rep[1])
        for z in range(rep[2])
    ]
    hydrogens = np.array(hh)
    hydrogens /= np.array(rep)

    return atoms_to_waters(oxygens, hydrogens, cell, partial_order=partial_order)


def shortest_distance(coord, cell):
    dmin = 1e99
    for i, j in it.combinations(coord, 2):
        d = i - j
        d -= np.floor(d + 0.5)
        d = d @ cell
        d = np.linalg.norm(d)
        if d < dmin:
            dmin = d
    return dmin


def density_in_g_cm3(nmol_water, cell):
    return 18 * nmol_water / (np.linalg.det(cell) * 1e-21 * 6.022e23)


def validate_ice_rules(dg: nx.DiGraph):
    logger = getLogger()
    valid = True
    for node in dg:
        in_degree = dg.in_degree(node)
        out_degree = dg.out_degree(node)
        if in_degree != 2 or out_degree != 2:
            valid = False
            logger.info(
                f"Node {node} has {dg.in_degree(node)} incoming edges"
                f" and {dg.out_degree(node)} outgoing edges"
            )
    return valid


def _closest_node_bruteforce(
    frac_pos: np.ndarray,
    lattice_sites: np.ndarray,
    cell: np.ndarray,
) -> int:
    """1 点 frac_pos に周期境界で最も近い格子点のインデックスを返す。総当り。"""
    p = np.asarray(frac_pos).reshape(3)
    best_j = 0
    best_d2 = np.inf
    for j in range(len(lattice_sites)):
        d = lattice_sites[j] - p
        d -= np.floor(d + 0.5)
        d_orth = d @ cell
        d2 = np.dot(d_orth, d_orth)
        if d2 < best_d2:
            best_d2 = d2
            best_j = j
    return best_j


def find_nearest_sites_pbc(
    frac_pos: Union[Tuple[float, float, float], np.ndarray],
    lattice_sites: np.ndarray,
    cell: np.ndarray,
) -> Union[int, List[int]]:
    """分数座標に最も近い格子サイトを返す。周期境界を考慮。pairs_iter で近傍を絞り、見つからなければ総当り。

    1 点のときは int、複数点のときは list[int] を返す。
    """
    p = np.atleast_2d(frac_pos)  # (1,3) または (N,3)
    n_query = len(p)

    @dataclass
    class Neighbor:
        distance: float
        node: int

    neighbors: dict[int, Neighbor] = {}
    for i, j, d in pl.pairs_iter(p, 0.5, cell, pos2=lattice_sites, distance=True):
        if i not in neighbors or d < neighbors[i].distance:
            neighbors[i] = Neighbor(distance=d, node=j)

    out = []
    for i in range(n_query):
        if i in neighbors:
            out.append(neighbors[i].node)
        else:
            out.append(_closest_node_bruteforce(p[i], lattice_sites, cell))

    return out[0] if n_query == 1 else out


def find_nearest_edges_pbc(
    frac_pos: Union[Tuple[float, float, float], np.ndarray],
    graph: nx.Graph,
    lattice_sites: np.ndarray,
    cell: np.ndarray,
) -> Union[Tuple[int, int], List[Tuple[int, int]]]:
    """分数座標に最も近い「辺（2つの格子サイトの中点）」を返す。周期境界を考慮。

    手順:
      1. まず find_nearest_sites_pbc で最寄りの格子サイト i を決める。
      2. i に隣接するノード j との中点 (i, j) のうち、target に最も近いものを選ぶ。

    1 点のときは (i, j)、複数点のときは list[(i, j)] を返す。
    """
    p = np.atleast_2d(frac_pos)  # (1,3) または (N,3)
    n_query = len(p)

    out: List[Tuple[int, int]] = []
    for q in p:
        # 最寄りサイト i を決める
        i_site = find_nearest_sites_pbc(q, lattice_sites, cell)
        # 隣接ノードとの中点のうち、もっとも近いものを選ぶ
        best_edge: Tuple[int, int] | None = None
        best_d2 = np.inf
        for j_site in graph.neighbors(i_site):
            midpoint = 0.5 * (lattice_sites[i_site] + lattice_sites[j_site])
            d = midpoint - q
            d -= np.floor(d + 0.5)
            d_orth = d @ cell
            d2 = float(np.dot(d_orth, d_orth))
            if d2 < best_d2:
                best_d2 = d2
                best_edge = (i_site, j_site)
        if best_edge is None:
            # i_site に隣接ノードがない場合はエラーにしておく
            raise ValueError(f"Node {i_site} has no neighbors in graph.")
        out.append(best_edge)

    return out[0] if n_query == 1 else out


# Candidates for ice XI
# {'P12_11', 'P2_12_12_1', 'Pca2_1', 'P1c1', 'Cmc2_1', 'C1c1', 'P12', 'Pbn2_1', 'Pna2_1'}
def operations(spaceg: str, origin: int = 0):
    symops = None
    if spaceg == "P12_11" or spaceg == "4":
        symops = symmetry_operators(
            """
            x,            y,            z
            -x,          1/2+y,         -z
        """
        )
    elif spaceg == "P2_12_12_1" or spaceg == "19":
        # http://img.chem.ucl.ac.uk/sgp/large/019a.htm
        if origin == 0:
            symops = symmetry_operators(
                """
                x,            y,            z
                1/2+x,        1/2-y,         -z
                -x,          1/2+y,        1/2-z
                1/2-x,         -y,          1/2+z
            """
            )
    elif spaceg == "Pca2_1" or spaceg == 29:
        symops = symmetry_operators(
            """
            x,            y,            z
            1/2-x,          y,          1/2+z
            1/2+x,         -y,            z
            -x,           -y,          1/2+z
        """
        )
    elif spaceg == "P1c1" or spaceg == "7":
        symops = symmetry_operators(
            """
            x,            y,            z
            x,           -y,          1/2+z
        """
        )
    elif spaceg == "Cmc2_1" or spaceg == "36":
        symops = symmetry_operators(
            """
            x,            y,            z
            -x,            y,            z
            x,           -y,          1/2+z
            -x,           -y,          1/2+z
        """,
            offsets=[("+0", "+0", "+0"), ("+1/2", "+1/2", "+0")],
        )
    elif spaceg == "C1c1" or spaceg == "9":
        symops = symmetry_operators(
            """
            x,            y,            z
            x,           -y,          1/2+z
        """,
            offsets=[("+0", "+0", "+0"), ("+1/2", "+1/2", "+0")],
        )
    elif spaceg == "P1" or spaceg == "1":
        symops = symmetry_operators(
            """
            x,            y,            z
        """
        )
    elif spaceg == "Pbn2_1":
        # there are two #33 (?)
        symops = symmetry_operators(
            """
            x,            y,            z
            1/2-x,        1/2+y,          z
            1/2+x,        1/2-y,        1/2+z
            -x,           -y,          1/2+z
        """
        )
    elif spaceg == "Pna2_1":
        # there are two #33 (?)
        symops = symmetry_operators(
            """
            x,            y,            z
            1/2-x,        1/2+y,        1/2+z
            1/2+x,        1/2-y,          z
            -x,           -y,          1/2+z
        """
        )
    return symops



# coding: utf-8
"""
Test for generating water positions from cage positions
"""


import genice2.lattices
from genice2 import FrankKasper
from genice2.valueparser import parse_cages
from genice2.cell import cellvectors


class Lattice(genice2.lattices.Lattice):
    def __init__(self):
        self.cages = """
        12 0.5 0.5 0.5
        14 0.5 0.0 -0.25
        14 0.0 0.25 0.5
        14 0.75 0.5 0.0
        14 0.5 0.0 0.25
        12 0.0 0.0 0.0
        14 0.25 0.5 0.0
        14 0.0 -0.25 0.5
        """

        self.cell = cellvectors(a=12.747893943706936,
                                b=12.747893943706936,
                                c=12.747893943706936)
        cagepos, cagetype = parse_cages(self.cages)
        self.lattice_sites = [w for w in FrankKasper.toWater(cagepos, self.cell)]
        self.coord = "relative"
        self.density = FrankKasper.estimate_density(
            self.lattice_sites, self.cell, 2.76)
        self.bondlen = 2.76 * 1.2



# ============================================================================
# New genice3.unitcell implementation (TODO: implement manually)
# ============================================================================

"""
Test for generating water positions from cage positions
"""

import genice3.unitcell
import numpy as np
from cif2ice import cellvectors


class UnitCell(genice3.unitcell.UnitCell):
    """
    __test3単位胞を定義するクラス。

    NOTE: This unitcell is not yet implemented.
    Please contact the maintainer or implement it manually.
    """

    def __init__(self, **kwargs):
        raise NotImplementedError(
            f"{self.__class__.__name__} is not yet implemented. "
            "This unitcell requires manual implementation. "
            "Please contact the maintainer or implement it manually. "
            f"Reason: lattice_sitesが定義されていないため"
        )


#!/usr/bin/python
"""
Usage: genice3 c0te
"""

from genice3.util import (
    atomdic,
    symmetry_operators,
    waters_and_pairs,
    density_in_g_cm3,
)
import genice3.unitcell
from logging import getLogger
import numpy as np
import networkx as nx
from cif2ice import cellvectors

desc = {
    "ref": {"C0": "Teeratchanan 2015"},
    "brief": "Filled ice C0 by Teeratchanan (Hydrogen-disordered.) (Positions of guests are supplied.)",
    "test": ({"options": "-r 2 2 2"},),
}


class UnitCell(genice3.unitcell.UnitCell):
    def __init__(self, **kwargs):
        logger = getLogger()

        # Ref. 2atom
        atoms = """
        O1 0.2342 0.4721 0.8019
        O2 0.7648 0.5306 0.2941
        Ne1 -0.0647 0.7868 0.7669
        """

        # Ref.
        # space group: P3_2
        symops = """
          x,            y,            z
         -y,x-y,z+2/3
         -x+y,-x,z+1/3
        """.replace(
            ",", " "
        )

        # Ref. 2cell
        a = 6.177 / 10.0  # nm
        c = 6.054 / 10.0  # nm
        C = 120.0

        cell = cellvectors(a, a, c, C=C)

        # helper routines to make from CIF-like data

        atomd = atomdic(atoms)
        # atoms = fullatoms(atomd, symmetry_operators(symops))
        sops = symmetry_operators(symops)

        # cagetype = []
        # cagepos = []
        # for atomname, pos in atoms:
        #     if atomname == "Ne1":
        #         self.cagetype.append(atomname)
        #         self.cagepos.append(pos)

        waters, pairs = waters_and_pairs(cell, atomd, sops)

        density = density_in_g_cm3(len(waters), cell)
        coord = "relative"
        super().__init__(
            cell=cell,
            lattice_sites=waters,
            density=density,
            coord=coord,
            bondlen=0.3,
            **kwargs,
        )



import numpy as np
import genice3.molecule


class Molecule(genice3.molecule.Molecule):
    """
    空のケージ用のプレースホルダー。
    """

    def __init__(self):
        # placeholder for empty cage

        sites = []
        labels = []
        name = "empty"
        is_water = False
        super().__init__(sites=sites, labels=labels, name=name, is_water=is_water)



desc = {
    "ref": {},
    "brief": "Conventional high-pressure ice VII.",
}

import genice3.unitcell
import numpy as np
import networkx as nx
from cif2ice import cellvectors


class UnitCell(genice3.unitcell.UnitCell):
    """
    ice7単位胞を定義するクラス。
    """

    def __init__(self, **kwargs):
        pairs_str = """
        0 4
        0 6
        1 4
        1 5
        1 6
        2 6
        3 4
        3 5
        3 6
        3 7
        4 2
        5 0
        5 2
        7 0
        7 1
        7 2
        8 12
        8 14
        9 12
        9 13
        9 14
        10 14
        11 12
        11 13
        11 14
        11 15
        12 10
        13 8
        13 10
        15 8
        15 9
        15 10
        """.split(
            "\n"
        )
        pairs = []
        for line in pairs_str:
            cols = line.split()
            if len(cols) == 2:
                pairs.append((int(cols[0]), int(cols[1])))

        graph = nx.Graph(pairs)

        waters = np.fromstring(
            """
        0.00 0.00 0.00
        0.50 0.50 0.00
        0.50 0.00 0.50
        0.00 0.50 0.50
        0.25 0.25 0.25
        0.75 0.75 0.25
        0.75 0.25 0.75
        0.25 0.75 0.75
        0.50 0.00 0.00
        0.00 0.50 0.00
        0.00 0.00 0.50
        0.50 0.50 0.50
        0.75 0.25 0.25
        0.25 0.75 0.25
        0.25 0.25 0.75
        0.75 0.75 0.75
        """,
            sep=" ",
        ).reshape(-1, 3)

        coord = "relative"

        # bondlen = 3

        density = 1.6

        cell = cellvectors(a=4.0, b=4.0, c=4.0)

        super().__init__(
            cell=cell,
            lattice_sites=waters,
            graph=graph,
            coord=coord,
            # bondlen=bondlen,
            density=density,
            **kwargs,
        )



# 分子の向きは違えど、おおよそ同じ形であることを検証する。

import sys
import numpy as np
from logging import basicConfig, INFO, getLogger, DEBUG
import networkx as nx
from dataclasses import dataclass


@dataclass
class Product:
    oxygens: np.ndarray
    graph: nx.Graph
    volume: float
    cell: np.ndarray


def genice2_generate_ice(ice, options):

    from genice2.genice import GenIce
    from genice2.plugin import Lattice, Format, Molecule

    lattice = Lattice(ice, **options)
    formatter = Format(
        "raw",
        stage=(
            1,  # replicated molecular positions
            2,  # replicated graph
            6,  # atomic positions of water
        ),
    )
    water = Molecule("physical_water")
    genice2 = GenIce(lattice).generate_ice(formatter, water=water)
    cell = genice2["repcell"]
    celli = np.linalg.inv(cell)
    volume = np.linalg.det(cell)
    oxygens = np.array([sites[0, :] for sites in genice2["mols"][0].positions]) @ celli
    return Product(oxygens=oxygens, graph=genice2["graph"], volume=volume, cell=cell)


def GenIce3_generate_ice(ice, options):
    from genice3.plugin import UnitCell
    from genice3.plugin import Molecule
    from genice3.genice import GenIce3
    from genice3.util import validate_ice_rules

    genice3 = GenIce3()
    genice3.unitcell = UnitCell(ice, **options)

    if not validate_ice_rules(genice3.digraph):
        # ice rulesに従わない
        if ice not in ("2D3", "iceM", "iceT2", "M"):
            raise ValueError(f"Violation of the ice rules in {ice} {options=}")

    waters = genice3.water_molecules(water_model=Molecule("physical_water"))
    cell = genice3.cell
    celli = np.linalg.inv(cell)
    volume = np.linalg.det(cell)
    oxygens = np.array([mol.sites[0] for mol in waters.values()]) @ celli
    # assert the volumes are similar
    return Product(oxygens=oxygens, graph=genice3.graph, volume=volume, cell=cell)


basicConfig(level=INFO)
logger = getLogger()
for ice in sys.argv[1:]:

    if ice in ("iceMd", "iceT", "CIF"):
        continue
    if ice == "xFAU":
        options = [dict(rep=4), dict(rep=8)]
    elif ice in ("one", "eleven"):
        options = [dict(layers="ccchchc"), dict(layers="ccchh")]
    elif ice == "11i":
        options = [dict(type=x) for x in range(1, 17)]
    elif ice == "bilayer":
        options = [dict(sw=0.1)]
    else:
        options = [dict()]

    for op in options:
        product2 = genice2_generate_ice(ice, op)
        product3 = GenIce3_generate_ice(ice, op)

        assert (
            abs(product2.volume - product3.volume) / product2.volume < 1e-6
        ), f"{product2.volume=}, {product3.volume=}"
        # 酸素原子位置を比較
        # 酸素の一対一対応を確認する(順序が変わっているかもしれないので。)
        import pairlist as pl

        pairs = [
            d
            for _, _, d in pl.pairs_iter(
                product2.oxygens,
                maxdist=0.1,
                cell=product2.cell,
                pos2=product3.oxygens,
                distance=True,
            )
            if d < 0.01
        ]
        logger.info(f"{min(pairs)=}, {max(pairs)=}")
        assert len(pairs) == len(
            product2.oxygens
        ), f"{len(pairs)=}, {len(product2.oxygens)=}"

        # これがなかなかに時間を食う。
        if "CRN" not in ice:
            assert nx.is_isomorphic(
                product2.graph, product3.graph
            ), f"{product2.graph.size()=}, {product3.graph.size()=}"

    # 水素位置の照合はしない。
    # bond_centers = []
    # for i, j in genice3.graph.edges():
    #     d = relative_O3[j] - relative_O3[i]
    #     d -= np.floor(d + 0.5)
    #     bond_centers.append(d / 2 + relative_O3[i])
    # bond_centers = np.array(bond_centers)

    # pairs = [
    #     d
    #     for x, y, d in pl.pairs_iter(
    #         relative_H2, maxdist=0.1, cell=cell2, pos2=bond_centers, distance=True
    #     )
    # ]
    # print(f"{min(pairs)=}, {max(pairs)=}")
    # assert len(pairs) == len(relative_H2), f"{pairs=}"

    # pairs = [
    #     d
    #     for x, y, d in pl.pairs_iter(
    #         relative_H3, maxdist=0.093, cell=cell2, pos2=bond_centers, distance=True
    #     )
    # ]
    # print(f"{min(pairs)=}, {max(pairs)=}")
    # assert len(pairs) == len(relative_H3), f"{len(pairs)=}, {len(relative_H3)=}"
